package Main;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;

import struct.sequence.ConfidenceEstimator;
import struct.sequence.LabelConfScore;
import struct.sequence.SequenceLabel;
import struct.sequence.SequencePrediction;
import struct.sequence.SequencePredictor;
import struct.types.Features;
import struct.types.Instance;
import cw_library.tools.MathTools;

public class CompareConfMeasures {
	tagger callingTagger;

	public CompareConfMeasures(tagger callingTagger) {
		this.callingTagger = callingTagger;
	}

	//
	// Go over entire list of words.
	// Each time a "relevant" word is encountered (an incorrect word) then
	// calculate "precision at" for this word rank and add it to sum of
	// precision-at values. Finally divide the sum of precision-at by the number
	// of incorrect words found.
	//
	private double calcMeanAveragePrecision(boolean[] words) {
		// go over all words in the array and search for incorrect words
		int totalIncorrectFound = 0, totalWordsChecked = 0;

		double precisionAtSum = 0;
		for (int idx = 0; idx < words.length; idx++) {
			totalWordsChecked++;

			if (words[idx] == false) {
				// Found incorrect word...
				totalIncorrectFound++;
				double precisionAt = (double) totalIncorrectFound
						/ (double) totalWordsChecked;
				precisionAtSum += precisionAt;
			}
		}// while

		return (precisionAtSum / totalIncorrectFound);
	}

	//
	// Fill in the "double[] precisionAt" with precision at K values as defined
	// by the precisionAtVals array
	//
	private void getPrecisionAtByWordsSorting(boolean[] words,
			int[] precisionAtVals, double[] precisionAtResults) {
		int numOfPrecisionAtValues = precisionAtVals.length;
		// Start with first "precision At" value
		int currPrecisionAtIndex = 0;
		int currPrecisionAtValue = precisionAtVals[currPrecisionAtIndex];

		// go over all words in the list and search for incorrect words
		int totalIncorrectFound = 0, totalWordsChecked = 0;

		for (int idx = 0; idx < words.length; idx++) {
			totalWordsChecked++;

			if (words[idx] == false) {
				// Found incorrect word...
				totalIncorrectFound++;
				if (totalIncorrectFound == currPrecisionAtValue) {
					// Reached the current "precision At" value, so store
					// current precision
					precisionAtResults[currPrecisionAtIndex] = (double) totalIncorrectFound
							/ (double) totalWordsChecked;
					// And search next prec-At value
					currPrecisionAtIndex++;
					if (currPrecisionAtIndex == numOfPrecisionAtValues) {
						break;
					}
					currPrecisionAtValue = precisionAtVals[currPrecisionAtIndex];
				}
			}
		}// while
	}

	//
	// Return an int[] that at each index 'i' holds the total count of incorrect
	// found in the first 'i' words
	//
	private void getRecallCountProgressByWordsSorting(boolean[] words,
			int stopAfter, double[] recallResults) {
		int idx;
		int totalIncorrectFound = 0;

		for (idx = 1; idx <= stopAfter; idx++) {
			if (words[idx - 1] == false) {
				totalIncorrectFound++;
			}
			recallResults[idx - 1] = totalIncorrectFound;
		}
	}

	private boolean areEqualPrediction(String[] pred1, String[] pred2) {
		if (pred1.length != pred2.length) {
			return false;
		}
		for (int i = 0; i < pred1.length; i++) {
			if (pred1[i].equals(pred2[i]) == false) {
				return false;
			}
		}
		return true;
	}

	//
	// Given the data, predict its tags by CW-Predictor by (1) K-Best Viterbi,
	// and (2) K-Draws.
	// Then evaluate the confidence in each word prediction in several methods
	// finally, sort the list of words according to the different confidence
	// estimation methods and see which method place the most incorrectly
	// labeled words at the beginning of the sorted list. More incorrect words
	// indicates a better "estimation" method.
	//	
	public void compareConfMeasures(Instance[] data,
			SequencePredictor predictor, String featureFile, int measureAt,
			String[] confEstimatorsNames) throws FileNotFoundException,
			IOException {

		int repeatCount = 10;
		System.out.println("Repetitions: " + repeatCount);

		int confBinsCount = 20;
		System.out.println("Confidence Bins count: " + confBinsCount);

		int estimatorsNum = confEstimatorsNames.length;
		boolean[][][] labelsSortedByConfidence = new boolean[estimatorsNum][repeatCount][];
		int[][][] binCorrect = new int[estimatorsNum][repeatCount][confBinsCount];
		int[][][] binIncorrect = new int[estimatorsNum][repeatCount][confBinsCount];
		int estimatorIdx;
		for (estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
				labelsSortedByConfidence[estimatorIdx][repIdx] = runConfMeasures(
						data, predictor, featureFile, measureAt,
						confEstimatorsNames[estimatorIdx],
						binCorrect[estimatorIdx][repIdx],
						binIncorrect[estimatorIdx][repIdx]);
			}

		}

		//
		// Now process retrieval results
		//
		processResultsIncorrectTagsDetection(measureAt, confEstimatorsNames,
				labelsSortedByConfidence);

		//
		// Now process confidence-bins results
		//
		processResultsConfidenceBins(measureAt, confEstimatorsNames,
				binCorrect, binIncorrect);
		processResultsConfidenceBins__full_Version(measureAt,
				confEstimatorsNames, binCorrect, binIncorrect);
	}

	//
	// bin[estimators][repetitions][bins-count]
	//
	private void processResultsConfidenceBins(int measureAt,
			String[] confEstimatorsNames, int[][][] binCorrect,
			int[][][] binIncorrect) {
		int estimatorsNum = binCorrect.length;
		int repeatCount = binCorrect[0].length;
		int binsCount = binCorrect[0][0].length;

		int[][] binCorrectSum = new int[estimatorsNum][binsCount];
		int[][] binIncorrectSum = new int[estimatorsNum][binsCount];

		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			for (int binIdx = 0; binIdx < binsCount; binIdx++) {
				for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
					binCorrectSum[estimatorIdx][binIdx] += binCorrect[estimatorIdx][repIdx][binIdx];
					binIncorrectSum[estimatorIdx][binIdx] += binIncorrect[estimatorIdx][repIdx][binIdx];
				}
			}
		}
		// divide by repetitions count
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			for (int binIdx = 0; binIdx < binsCount; binIdx++) {
				binCorrectSum[estimatorIdx][binIdx] /= repeatCount;
				binIncorrectSum[estimatorIdx][binIdx] /= repeatCount;
			}
		}

		System.out.println();
		System.out.println("Confidence bins results: " + binsCount
				+ " Bins. Sum over " + repeatCount + " runs.");

		//
		// Print the correct\incorrect counts per bin
		//
		System.out.println();

		// Print the measure types names
		System.out.print("Correct count:" + "\t");
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(confEstimatorsNames[estimatorIdx] + "\t");
		}
		System.out.println();

		// Print the values
		for (int binIdx = 0; binIdx < binsCount; binIdx++) {
			System.out.print(binIdx + "\t");
			for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
				System.out.print(binCorrectSum[estimatorIdx][binIdx] + "\t");
			}
			System.out.println();
		}
		System.out.println();

		System.out.print("Incorrect count:" + "\t");
		System.out.println();

		// Print the values
		for (int binIdx = 0; binIdx < binsCount; binIdx++) {
			System.out.print(binIdx + "\t");
			for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
				System.out.print(binIncorrectSum[estimatorIdx][binIdx] + "\t");
			}
			System.out.println();
		}
		System.out.println();
	}

	private void processResultsConfidenceBins__full_Version(int measureAt,
			String[] confEstimatorsNames, int[][][] binCorrect,
			int[][][] binIncorrect) {
		int estimatorsNum = binCorrect.length;
		int repeatCount = binCorrect[0].length;
		int binsCount = binCorrect[0][0].length;

		//
		// Calc Rooted-Weighted-Square-Error (aka standard error) between the
		// bin's actual accuracy and expected accuracy by confidence estimate
		//
		double[][] standardErrPerRepetition = new double[estimatorsNum][repeatCount];
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
				standardErrPerRepetition[estimatorIdx][repIdx] = calcStandardError(
						binCorrect[estimatorIdx][repIdx],
						binIncorrect[estimatorIdx][repIdx]);
			}

		}

		//
		// Average standard-error over all repetitions
		//
		double[] standardErr = new double[estimatorsNum];
		double[] standardErrStddev = new double[estimatorsNum];
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			standardErr[estimatorIdx] = MathTools
					.calculateMean(standardErrPerRepetition[estimatorIdx]);
			standardErrStddev[estimatorIdx] = MathTools
					.calculateSTDev(standardErrPerRepetition[estimatorIdx]);
		}

		//
		// Count percentage of items in bin
		//
		double percentageOfItemsInBin[][] = new double[estimatorsNum][binsCount];
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			//
			// Count total items in all bins per estimator type and not for just
			// any estimator because the bins method may not be valid for some
			// estimator
			//
			int totalItems = 0;
			for (int binIdx = 0; binIdx < binsCount; binIdx++) {
				totalItems += binCorrect[estimatorIdx][0][binIdx]
						+ binIncorrect[estimatorIdx][0][binIdx];
			}

			//
			// Now per bin count items across the multiple repetitions
			//
			for (int binIdx = 0; binIdx < binsCount; binIdx++) {
				double totalPercentageOfItemsInBin = 0;
				for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
					int itemsInBin = (binCorrect[estimatorIdx][repIdx][binIdx] + binIncorrect[estimatorIdx][repIdx][binIdx]);
					totalPercentageOfItemsInBin += ((double) itemsInBin / (double) totalItems);
				}

				// And average percentage
				percentageOfItemsInBin[estimatorIdx][binIdx] = totalPercentageOfItemsInBin
						/ repeatCount;
			}
		}

		//
		// Calculate average accuracy of each bin
		//
		double avgAccuracyOfBin[][] = new double[estimatorsNum][binsCount];
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			for (int binIdx = 0; binIdx < binsCount; binIdx++) {
				double totAccuracyOfBin = 0;
				for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
					int itemsInBin = (binCorrect[estimatorIdx][repIdx][binIdx] + binIncorrect[estimatorIdx][repIdx][binIdx]);
					if (itemsInBin == 0) {
						continue;
					}
					int correctItemsInBin = binCorrect[estimatorIdx][repIdx][binIdx];
					double accuracy = (double) correctItemsInBin
							/ (double) itemsInBin;
					totAccuracyOfBin += accuracy;
				}

				// And average accuracy
				avgAccuracyOfBin[estimatorIdx][binIdx] = totAccuracyOfBin
						/ repeatCount;
			}
		}

		System.out.println();
		System.out.println("Confidence bins results:");

		//
		// Print the standard-error results
		//
		System.out.println();
		System.out.println("Standard Error results:");

		// Print the measure types names
		System.out.print("Std-Err:" + "\t");
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(confEstimatorsNames[estimatorIdx] + "\t");
		}
		System.out.println();

		// Print the values
		DecimalFormat df = new DecimalFormat();
		df.setMaximumFractionDigits(5);
		System.out.print("\t");
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(df.format(standardErr[estimatorIdx]) + "\t");
		}
		System.out.println();
		System.out.print("Stddev:" + "\t");
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(df.format(standardErrStddev[estimatorIdx]) + "\t");
		}
		System.out.println();

		//
		// Print the percentage of items and average accuracy per bin
		//
		System.out.println();
		System.out.println("Per bin accuracy and percentage of items:");

		System.out.print("Accuracy" + "\t");
		// Print the measure types names
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(confEstimatorsNames[estimatorIdx] + "\t");
		}
		System.out.println();

		// Print the values
		df.setMaximumFractionDigits(5);
		for (int binIdx = 0; binIdx < binsCount; binIdx++) {
			double binConfidence = (double) binIdx / (double) binsCount;
			System.out.print(df.format(binConfidence) + "\t");
			for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
				System.out.print(df
						.format(avgAccuracyOfBin[estimatorIdx][binIdx])
						+ "\t");
			}
			System.out.println();
		}

		System.out.println();
		System.out.print("Precent" + "\t");
		// Print the measure types names
		for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			System.out.print(confEstimatorsNames[estimatorIdx] + "\t");
		}
		System.out.println();

		// Print the values
		df.setMaximumFractionDigits(5);
		for (int binIdx = 0; binIdx < binsCount; binIdx++) {
			double binConfidence = (double) binIdx / (double) binsCount;
			System.out.print(df.format(binConfidence) + "\t");
			for (int estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
				System.out.print(df
						.format(percentageOfItemsInBin[estimatorIdx][binIdx])
						+ "\t");
			}
			System.out.println();
		}

	}

	//
	// Calc Rooted-Weighted-Square-Error (aka standard error) between the
	// bin's actual accuracy and expected accuracy by confidence estimate
	//
	private double calcStandardError(int[] binCorrect, int[] binIncorrect) {
		int binsCount = binCorrect.length;

		// The difference in expected accuracy between successive bins: 1/K
		double binExpectedAccuracyInterval = ((double) 1 / binsCount);
		double initExpectedAccuracyInterval = binExpectedAccuracyInterval / 2;

		int totalItems = 0;
		double totalWgtSqErr = 0;
		double binExpectedAccuracy = initExpectedAccuracyInterval;
		for (int binIdx = 0; binIdx < binsCount; binIdx++, binExpectedAccuracy += binExpectedAccuracyInterval) {
			if (binIncorrect[binIdx] < 0) {
				// this estimation method is invalid for confidence bin as
				// confidence score may be out of [0,1]
				return -1;

			}

			int totItemsInBin = binCorrect[binIdx] + binIncorrect[binIdx];
			if (totItemsInBin == 0) {
				// some bins may be empty
				continue;
			}

			totalItems += totItemsInBin;
			double binAccuracy = (double) binCorrect[binIdx]
					/ (double) totItemsInBin;
			double accuracyErr = binAccuracy - binExpectedAccuracy;
			double weightedSquareErr = accuracyErr * accuracyErr
					* totItemsInBin;
			totalWgtSqErr += weightedSquareErr;
		}

		double meanSquareError = totalWgtSqErr / totalItems;
		double standardError = Math.sqrt(meanSquareError);
		return standardError;
	}

	public void processResultsIncorrectTagsDetection(int measureAt,
			String[] confEstimatorsNames, boolean[][][] labelsSortedByConfidence) {

		int repeatCount = labelsSortedByConfidence[0].length;
		int estimatorsNum = confEstimatorsNames.length;
		int estimatorIdx;

		//
		// Now process results
		//
		int totalCorrect = 0;
		int totalIncorrect = 0;
		for (int i = 0; i < labelsSortedByConfidence[0][0].length; i++) {
			if (labelsSortedByConfidence[0][0][i]) {
				totalCorrect++;
			} else {
				totalIncorrect++;
			}
		}
		System.out.println("Total correct: " + totalCorrect + ", Incorrect: "
				+ totalIncorrect + ", Total: "
				+ (totalCorrect + totalIncorrect));

		int stopAfterAll = totalCorrect + totalIncorrect;
		int stopAfter = Math.min(measureAt, stopAfterAll);
		// Max possibel recalls is the min between #of incorrects and #of
		// elements to rank
		int possibleRecalls = Math.min(totalIncorrect, stopAfter);

		// Results tables
		int maxMeasuresTypes = 100;
		double[][] resultsTable = new double[maxMeasuresTypes][];
		String[] measuresNames = new String[maxMeasuresTypes];

		// "Precision at" values
		// This means - when reaching recall of K, what is the precision.
		int[] precisionAtVals = { 100, 200, 300, 400, 500, 600, 700, 800,
				900, 1000, 1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800,
				1900, 2000 };
		int numOfPrecisionAtVals = precisionAtVals.length;
		double[][] precisionAtResTable = new double[maxMeasuresTypes][numOfPrecisionAtVals];

		// "Average Precision" values
		// This means - going over all the words, when incorrect word is
		// encountered calculate the current precision and then calculate the
		// mean of all the precision values calculated
		double[] mapResTable = new double[maxMeasuresTypes];
		double[] mapStddevResTable = new double[maxMeasuresTypes];

		//
		// Go over all words sortings and calculate results
		//
		for (estimatorIdx = 0; estimatorIdx < estimatorsNum; estimatorIdx++) {
			measuresNames[estimatorIdx] = confEstimatorsNames[estimatorIdx];

			// allocate temp restuls arrays for averaging over repetitions
			double[][] recallCountTable = new double[repeatCount][stopAfter];
			double[][] precAtResTable = new double[repeatCount][numOfPrecisionAtVals];
			double[] avgPrecTable = new double[repeatCount];
			// Go over all repetitions
			for (int repIdx = 0; repIdx < repeatCount; repIdx++) {
				boolean[] curLabelsSortedByConfidence = labelsSortedByConfidence[estimatorIdx][repIdx];
				getRecallCountProgressByWordsSorting(
						curLabelsSortedByConfidence, stopAfter,
						recallCountTable[repIdx]);
				getPrecisionAtByWordsSorting(curLabelsSortedByConfidence,
						precisionAtVals, precAtResTable[repIdx]);
				avgPrecTable[repIdx] = calcMeanAveragePrecision(curLabelsSortedByConfidence);
			}

			resultsTable[estimatorIdx] = MathTools
					.calculateArrayOfMean(recallCountTable);
			precisionAtResTable[estimatorIdx] = MathTools
					.calculateArrayOfMean(precAtResTable);
			mapResTable[estimatorIdx] = MathTools.calculateMean(avgPrecTable);
			mapStddevResTable[estimatorIdx] = MathTools
					.calculateSTDev(avgPrecTable);
		}

		//
		// Print the "recall at i" results
		//
		int numOfMeasureTypes = estimatorsNum;
		int measureTypeIdx;

		System.out.println("Recall at index i:");
		// Print the measure types names
		System.out.print("Method:" + "\t");
		for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
			System.out.print(measuresNames[measureTypeIdx] + "\t");
		}
		System.out.println();
		// Print the values
		DecimalFormat df = new DecimalFormat();
		df.setMaximumFractionDigits(5);
		for (int resIdx = 0; resIdx < stopAfter; resIdx++) {
			System.out.print("\t");
			for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
				System.out.print(df
						.format(resultsTable[measureTypeIdx][resIdx])
						+ "\t");
			}
			System.out.println();
		}

		//
		// Print the "Precision At" results
		//
		System.out.println();
		System.out.println("Precision at results:");

		System.out.print("PrecAt" + "\t");
		// Print the measure types names
		for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
			System.out.print(measuresNames[measureTypeIdx] + "\t");
		}
		System.out.println();

		// Print the values
		df.setMaximumFractionDigits(5);
		for (int resIdx = 0; resIdx < numOfPrecisionAtVals; resIdx++) {
			System.out.print(precisionAtVals[resIdx] + "\t");
			for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
				System.out.print(df
						.format(precisionAtResTable[measureTypeIdx][resIdx])
						+ "\t");
			}
			System.out.println();
		}

		//
		// Print the AP results
		//
		System.out.println();
		System.out.println("Average Precision results:");

		// Print the measure types names
		System.out.print("Avg-Prec:" + "\t");
		for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
			System.out.print(measuresNames[measureTypeIdx] + "\t");
		}
		System.out.println();

		// Print the values
		df.setMaximumFractionDigits(5);
		System.out.print("\t");
		for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
			System.out.print(df.format(mapResTable[measureTypeIdx]) + "\t");
		}
		System.out.println();
		System.out.print("Stddev:" + "\t");
		for (measureTypeIdx = 0; measureTypeIdx < numOfMeasureTypes; measureTypeIdx++) {
			System.out.print(df.format(mapStddevResTable[measureTypeIdx])
					+ "\t");
		}
		System.out.println();
	}

	LinkedList<LabelConfScore> allWordsScoresList;

	public boolean[] runConfMeasures(Instance[] data,
			SequencePredictor predictor, String featureFile, int measureAt,
			String confEstimatorName, int[] binCorrect, int[] binIncorrect)
			throws FileNotFoundException, IOException {
		ObjectInputStream in = (featureFile != null) ? new ObjectInputStream(
				new FileInputStream(featureFile)) : null;

		// Table of all the words from all the sentences
		LabelConfScore[][] allWordsScores = new LabelConfScore[data.length][];

		ConfidenceEstimator confEstimator = ConfidenceEstimator
				.resolveConfidenceEstimatorByName(confEstimatorName);
		// Go over all instances and collect statistics
		for (int instIdx = 0; instIdx < data.length; instIdx++) {

			Instance inst = data[instIdx];
			Features feats;
			feats = (in != null) ? data[instIdx].getFeatures(in)
					: data[instIdx].getFeatures();

			//
			// predict the labels sequence and for each tag mark as correct or
			// not
			//
			SequencePrediction seqPrediction = predictor.decode(inst, feats);

			String[] act_spans = ((SequenceLabel) inst.getLabel()).tags;
			String[] pred_spans = ((SequenceLabel) seqPrediction.getBestLabel()).tags;

			allWordsScores[instIdx] = new LabelConfScore[act_spans.length];
			int wordIdx;
			for (wordIdx = 0; wordIdx < act_spans.length; wordIdx++) {
				LabelConfScore wordScore = new LabelConfScore();
				allWordsScores[instIdx][wordIdx] = wordScore;

				// word is correct or not
				wordScore.isCorrect = (pred_spans[wordIdx]
						.equals(act_spans[wordIdx]));
			}

			//
			// Now assign the confidence score to the words
			//
			confEstimator.estimateInst(inst, feats, allWordsScores[instIdx],
					predictor);
		}
		if (in != null)
			in.close();

		//
		// Now from all the arrays of word-stats create one large lists
		//
		allWordsScoresList = new LinkedList<LabelConfScore>();
		int idx;
		for (idx = 0; idx < allWordsScores.length; idx++) {
			if (allWordsScores[idx] == null || allWordsScores[idx].length == 0) {
				// a skipped sentence (short)
				continue;
			}
			allWordsScoresList.addAll(Arrays.asList(allWordsScores[idx]));
		}

		//
		// Randomly shuffle the list then sort it by score
		//
		Collections.shuffle(allWordsScoresList);
		Collections.shuffle(allWordsScoresList);
		Collections.sort(allWordsScoresList, new CompareByScore());

		//
		// Create the sorted array of correct\incorrect words
		//
		boolean[] sortedWords = new boolean[allWordsScoresList.size()];
		int arrIdx;
		Iterator<LabelConfScore> iter = allWordsScoresList.iterator();
		LabelConfScore wordScore;

		arrIdx = 0;
		while (iter.hasNext()) {
			wordScore = iter.next();
			sortedWords[arrIdx] = wordScore.isCorrect;
			arrIdx++;
		}

		//
		// Now sort to words to confidence bins
		//
		int binsCount = binCorrect.length;
		LabelConfScore wordData;
		iter = allWordsScoresList.iterator();
		while (iter.hasNext()) {
			wordData = iter.next();

			// Correct tiny numeric glitches (for Gamma method)
			double err = 0.001;
			if (0 - err < wordData.score && wordData.score < 0) {
				wordData.score = 0;
			}
			if (1 < wordData.score && wordData.score < 1 + err) {
				wordData.score = 1;
			}

			if (wordData.score < 0 || wordData.score > 1) {
				//
				// Some conf-estimation methods (Delta) are not between [0,1], so the
				// conf-bins are not relevant...
				// Mark with '-1' in incorrect bin.
				//
				for (int binIdx = 0; binIdx < binsCount; binIdx++) {
					binCorrect[binIdx] = 0;
					binIncorrect[binIdx] = -1;
				}
				break;
			}

			int binIdx = (int) (wordData.score * binsCount);
			if (binIdx == binsCount) {
				// In case of perfect score
				binIdx--;
			}
			if (wordData.isCorrect) {
				binCorrect[binIdx]++;
			} else {
				binIncorrect[binIdx]++;
			}
		}

		return sortedWords;
	}

	class CompareByScore implements Comparator<LabelConfScore> {
		public CompareByScore() {
			super();
		}

		public int compare(LabelConfScore o1, LabelConfScore o2) {
			if (o1.score < o2.score)
				return -1;
			if (o2.score < o1.score)
				return 1;
			return 0;
		}
	}
}
