package Main;

import java.text.DecimalFormat;

import struct.sequence.ChunkerDataManager;
import struct.sequence.SequenceDataManager;
import struct.sequence.SequenceEvaluationAveragedResult;
import struct.sequence.SequenceEvaluationResult;
import cw_library.tools.MathTools;

public class CrossValidation {

	tagger callingTagger;

	public CrossValidation(tagger callingTagger) {
		this.callingTagger = callingTagger;
	}

	public void crossValidation(String fullFileName, String posNegResFileName,
			int cvFolds, int numTrainingIters, boolean bAvgParams, String predictByCommittee)
			throws Exception {

		TrainAndTest trainAndTest = new TrainAndTest(callingTagger);

		int lastIndexOfSlash;
		// Handle both Windows and Linux: ..\file or ../file
		lastIndexOfSlash = fullFileName.lastIndexOf("/");
		if (lastIndexOfSlash == -1)
			lastIndexOfSlash = fullFileName.lastIndexOf("\\");

		String filesPath = fullFileName.substring(0, lastIndexOfSlash + 1);
		String newTrainFileName = "train.txt";
		String newTestFileName = "eval.txt";
		String newModelFileName = "model.txt";

		// Cross validation.
		SequenceDataManager manager = callingTagger.resolveDataManagerByArgs();
		manager.splitDataFileToTrainAndEval(fullFileName, cvFolds, filesPath,
				newTrainFileName, newTestFileName);

		SequenceEvaluationResult[] evalResults = new SequenceEvaluationResult[cvFolds];
		for (int i = 0; i < evalResults.length; i++) {
			evalResults[i] = new SequenceEvaluationResult();
		}

		// At first iteration - create new file
		boolean appendToResFile = false;

		// Do the N folds
		for (int foldIdx = 0; foldIdx < cvFolds; foldIdx++) {
			System.out.println("Cross validation fold #" + foldIdx);

			String prefix = filesPath + String.valueOf(foldIdx) + "_";
			System.out.println("Files prefix: " + prefix);
			String iterTrainFileName = prefix + newTrainFileName;
			String iterTestFileName = prefix + newTestFileName;
			String iterModelFileName = prefix + newModelFileName;

			//
			// Train
			//
			trainAndTest.runTrain(iterTrainFileName, iterModelFileName,
					numTrainingIters, bAvgParams);

			//
			// Test
			//
			trainAndTest.runTest(iterTestFileName, iterModelFileName,
					evalResults[foldIdx], posNegResFileName, appendToResFile, predictByCommittee);

			// At next iterations - append to file
			appendToResFile = true;
		}

		// Calculate means and print the results
		boolean shouldPrint = true;
		calculateMeanAndStdDev(evalResults, shouldPrint);
	}

	//
	// Average k-results over the cv-folds:
	// input: results[cv-folds][k-results]
	// output: results[k-average-results]
	//
	public SequenceEvaluationAveragedResult[] calculateMeanAndStdDev(
			SequenceEvaluationResult[][] evalResults) {
		int cvFolds = evalResults.length;
		int numOfResultsPerFold = evalResults[0].length;
		SequenceEvaluationAveragedResult[] averagedResults = new SequenceEvaluationAveragedResult[numOfResultsPerFold];

		for (int resultIdx = 0; resultIdx < numOfResultsPerFold; resultIdx++) {
			SequenceEvaluationResult[] foldResults = new SequenceEvaluationResult[cvFolds];
			for (int foldIdx = 0; foldIdx < cvFolds; foldIdx++) {
				foldResults[foldIdx] = evalResults[foldIdx][resultIdx];
			}
			boolean shouldPrint = false;
			averagedResults[resultIdx] = calculateMeanAndStdDev(foldResults,
					shouldPrint);
		}

		return averagedResults;
	}

	public SequenceEvaluationAveragedResult calculateMeanAndStdDev(
			SequenceEvaluationResult[] evalResults, boolean bPrint) {
		int numFolds = evalResults.length;
		double[] prec = new double[numFolds];
		double[] recall = new double[numFolds];
		double[] fMeas = new double[numFolds];
		for (int foldIdx = 0; foldIdx < numFolds; foldIdx++) {
			prec[foldIdx] = evalResults[foldIdx].precision;
			recall[foldIdx] = evalResults[foldIdx].recall;
			fMeas[foldIdx] = evalResults[foldIdx].fMeasure;
		}
		double precMean = MathTools.calculateMean(prec);
		double precStddev = MathTools.calculateSTDev(prec, precMean);
		double recallMean = MathTools.calculateMean(recall);
		double recallStddev = MathTools.calculateSTDev(recall, recallMean);
		double fmeasMean = MathTools.calculateMean(fMeas);
		double fmeasStddev = MathTools.calculateSTDev(fMeas, fmeasMean);

		SequenceEvaluationAveragedResult averagedResult = new SequenceEvaluationAveragedResult();
		averagedResult.setValues(precMean, recallMean, fmeasMean);
		averagedResult.setStddevValues(precStddev, recallStddev, fmeasStddev);

		if (bPrint) {
			DecimalFormat nf = new DecimalFormat();
			nf.setMaximumFractionDigits(5);

			// Print the results of each fold
			System.out.println();
			System.out.println("Results of each fold:");
			// System.out.println("Prec\tRecall\tF-Meas");
			// for (int foldIdx = 0; foldIdx < numFolds; foldIdx++) {
			// System.out.print(nf.format(prec[foldIdx]) + "\t");
			// System.out.print(nf.format(recall[foldIdx]) + "\t");
			// System.out.print(nf.format(fMeas[foldIdx]) + "\t");
			// System.out.println();
			// }
			SequenceEvaluationResult.printResultsArray(evalResults);

			System.out.println();
			// Print the results averaged over all folds
			System.out.println("Results averaged over " + numFolds
					+ " fold CV:");
			// System.out.println("Prec, Recall and F-Meas Mean and StdDev:");
			// System.out.println(nf.format(precMean) + "\t" +
			// nf.format(precStddev));
			// System.out.println(nf.format(recallMean) + "\t"
			// + nf.format(recallStddev));
			// System.out
			// .println(nf.format(fmeasMean) + "\t" + nf.format(fmeasStddev));

			SequenceEvaluationAveragedResult[] averagedResults = new SequenceEvaluationAveragedResult[1];
			averagedResults[0] = averagedResult;
			SequenceEvaluationAveragedResult.printResultsArray(averagedResults);
		}

		return averagedResult;
	}

	public void crossValidationIncrementalTraining(String fullFileName,
			String posNegResFileName, int cvFolds, boolean bAvgParams,
			int maxNumOfTrainingIters, String predictByCommittee) throws Exception {

		TrainAndTest trainAndTest = new TrainAndTest(callingTagger);

		int lastIndexOfSlash;
		// Handle both Windows and Linux: ..\file or ../file
		lastIndexOfSlash = fullFileName.lastIndexOf("/");
		if (lastIndexOfSlash == -1)
			lastIndexOfSlash = fullFileName.lastIndexOf("\\");

		String filesPath = fullFileName.substring(0, lastIndexOfSlash + 1);
		String newTrainFileName = "train.txt";
		String newTestFileName = "eval.txt";
		String newModelFileName = "model.txt";

		// Cross validation.
		SequenceDataManager manager = callingTagger.resolveDataManagerByArgs();
		manager.splitDataFileToTrainAndEval(fullFileName, cvFolds, filesPath,
				newTrainFileName, newTestFileName);

		SequenceEvaluationResult[][] evalResults = new SequenceEvaluationResult[cvFolds][maxNumOfTrainingIters];
		for (int i = 0; i < evalResults.length; i++) {
			evalResults[i] = new SequenceEvaluationResult[maxNumOfTrainingIters];
			for (int j = 0; j < evalResults[i].length; j++) {
				evalResults[i][j] = new SequenceEvaluationResult();
			}
		}

		// At first iteration - create new file
		boolean appendToResFile = false;

		// Do the N folds
		for (int foldIdx = 0; foldIdx < cvFolds; foldIdx++) {
			System.out.println("Cross validation fold #" + foldIdx);

			String prefix = filesPath + String.valueOf(foldIdx) + "_";
			System.out.println("Files prefix: " + prefix);
			String iterTrainFileName = prefix + newTrainFileName;
			String iterTestFileName = prefix + newTestFileName;
			String iterModelFileName = prefix + newModelFileName;

			// run training using 1,2,...maxIters iterations and evaluate
			// performance after each iteration
			trainAndTest.incrementalIterationsTrainAndTest(iterTrainFileName,
					iterModelFileName, iterTestFileName, maxNumOfTrainingIters,
					bAvgParams, evalResults[foldIdx], posNegResFileName,
					appendToResFile, predictByCommittee);

			// At next iterations - append to file
			appendToResFile = true;
		}

		// Calculate means of the results
		SequenceEvaluationAveragedResult[] averagedResults = calculateMeanAndStdDev(evalResults);
		System.out.println();
		// Print the results averaged over all folds
		System.out.println("Results averaged over " + cvFolds + " fold CV:");
		SequenceEvaluationAveragedResult.printResultsArray(averagedResults);
	}

	private SequenceEvaluationResult calcAveragedResult(
			SequenceEvaluationResult[] resultsArr) {
		int numItems = resultsArr.length;
		double[] prec = new double[numItems];
		double[] recall = new double[numItems];
		double[] fMeas = new double[numItems];
		for (int itemIdx = 0; itemIdx < numItems; itemIdx++) {
			prec[itemIdx] = resultsArr[itemIdx].precision;
			recall[itemIdx] = resultsArr[itemIdx].recall;
			fMeas[itemIdx] = resultsArr[itemIdx].fMeasure;
		}
		double precMean = MathTools.calculateMean(prec);
		double recallMean = MathTools.calculateMean(recall);
		double fmeasMean = MathTools.calculateMean(fMeas);

		SequenceEvaluationResult avgRes = new SequenceEvaluationResult();
		avgRes.setValues(precMean, recallMean, fmeasMean);
		return avgRes;
	}
}
