package Main;

import struct.alg.OnlineLearner;
import struct.alg.OnlineUpdator;
import struct.sequence.CWSequencePredictor;
import struct.sequence.SequenceDataManager;
import struct.sequence.SequenceEvaluationResult;
import struct.sequence.SequenceEvaluator;
import struct.sequence.SequenceInstance;
import struct.sequence.SequencePredictor;
import struct.sequence.SequencePredictorByMajorityVote;
import cw_library.tools.MathTools;

public class TrainAndTest {
	tagger _tagger;

	public TrainAndTest(tagger callingTagger) {
		this._tagger = callingTagger;
	}

	public void runTrain(String _trainFileName, String _modelFileName,
			int numTrainingIters, boolean bAvgParams) throws Exception {
		SequenceDataManager manager = _tagger.resolveDataManagerByArgs();

		SequenceInstance[] trainingData = genTrainingData(_trainFileName,
				manager);
		trainOnData(_trainFileName, _modelFileName, numTrainingIters,
				bAvgParams, manager, trainingData);
	}

	public SequenceInstance[] genTrainingData(String _trainFileName,
			SequenceDataManager manager) throws Exception {

		boolean createUnsupportedFeatures = false;
		manager.createAlphabets(_trainFileName, createUnsupportedFeatures);
		manager.closeAlphabets();

		int numFeats = manager.dataAlphabet.size();
		System.out.println("Feature Dimension: " + numFeats);

		boolean createFeaturesFile = true;
		boolean createCrfFile = false;
		SequenceInstance[] trainingData = manager.readData(_trainFileName,
				createFeaturesFile, createCrfFile);
		return trainingData;
	}

	public void trainOnData(String _trainFileName, String _modelFileName,
			int numTrainingIters, boolean bAvgParams,
			SequenceDataManager manager, SequenceInstance[] trainingData)
			throws Exception {
		int numFeats = manager.dataAlphabet.size();

		OnlineLearner learner = new OnlineLearner();

		SequencePredictor predictor = _tagger
				.resolvePredictorByArgs(numFeats);
		OnlineUpdator updator = _tagger.resolveUpdatorByArgs();

		learner.train(trainingData, updator, predictor, numTrainingIters,
				_trainFileName + ".feats", bAvgParams);

		predictor.saveModel(_modelFileName);

		//
		// Calc model weights mean and StdDev
		//
		double[] weightsVector = predictor.weights;
		double weightsMean = MathTools.calculateMean(weightsVector);
		double weightsStddev = MathTools.calculateSTDev(weightsVector,
				weightsMean);
		System.out.println("Weights values mean and StdDev: " + weightsMean
				+ ", " + weightsStddev);
		double weightsAbsMean = MathTools.calculateAbsMean(weightsVector);
		double weightsAbsStddev = MathTools.calculateAbsSTDev(weightsVector,
				weightsMean);
		System.out.println("Weights Abs values mean and StdDev: "
				+ weightsAbsMean + ", " + weightsAbsStddev);

		//
		// If CW type of alg, calc sigma mean and StdDev
		//		
		if (predictor instanceof CWSequencePredictor) {
			double[] sigma = ((CWSequencePredictor) predictor).sigma;
			double sigmaMean = MathTools.calculateMean(sigma);
			double sigmaStddev = MathTools.calculateSTDev(sigma, sigmaMean);
			System.out.println("Sigma mean and StdDev: " + sigmaMean + ", "
					+ sigmaStddev);
			double stddevMean = MathTools.calculateMeanOfSqrts(sigma);
			double stddevStddev = MathTools.calculateSTDevOfSqrts(sigma, stddevMean);
			System.out.println("Stddev mean and StdDev: " + stddevMean + ", "
					+ stddevStddev);
		}
	}

	public void runTest(String _testFileName, String _modelFileName,
			SequenceEvaluationResult evalResults, String posNegResFile,
			boolean appendToResFile, String predictByCommittee)
			throws Exception {

		SequencePredictor predictor = _tagger.resolvePredictorByArgs(0);
		predictor.loadModel(_modelFileName);
		if (predictByCommittee != null) {
			predictor = new SequencePredictorByMajorityVote(predictor,
					predictByCommittee);
		}

		SequenceDataManager manager = _tagger.resolveDataManagerByArgs();
		SequenceInstance[] evalData = genEvalData(_testFileName,
				_modelFileName, manager);
		evalOnData(predictor, _testFileName, _modelFileName, evalResults,
				posNegResFile, appendToResFile, manager, evalData);
	}

	public SequenceInstance[] genEvalData(String _testFileName,
			String _modelFileName, SequenceDataManager manager)
			throws Exception {
		manager.dataAlphabet = SequenceInstance.getDataAlphabet();
		manager.tagAlphabet = SequenceInstance.getTagAlphabet();
		manager.closeAlphabets();

		int numFeats = manager.dataAlphabet.size();
		System.out.println("Feature Dimension: " + numFeats);

		boolean createFeaturesFile = true;
		boolean createCrfFile = false;
		SequenceInstance[] evalData = manager.readData(_testFileName,
				createFeaturesFile, createCrfFile);
		return evalData;
	}

	public void evalOnData(SequencePredictor predictor, String _testFileName,
			String _modelFileName, SequenceEvaluationResult evalResults,
			String posNegResFile, boolean appendToResFile,
			SequenceDataManager manager, SequenceInstance[] evalData)
			throws Exception {

		SequenceEvaluator eval = new SequenceEvaluator(manager.getTagAlphabet());

		eval.evaluate(evalData, predictor, _testFileName + ".feats",
				evalResults, posNegResFile, appendToResFile);
	}

	public void incrementalIterationsTrainAndTest(String _trainFileName,
			String _modelFileName, String _testFileName,
			int maxNumOfTrainingIters, boolean bAvgParams,
			SequenceEvaluationResult[] evalResults, String posNegResFile,
			boolean appendToResFile, String predictByCommittee)
			throws Exception {

		//
		// evalResults array may be non-null if it is required by caller, or
		// null otherwise
		//
		if (evalResults == null) {
			evalResults = new SequenceEvaluationResult[maxNumOfTrainingIters];
			for (int idx = 0; idx < evalResults.length; idx++) {
				evalResults[idx] = new SequenceEvaluationResult();
			}
		}

		// Gen data manger and training data once
		SequenceDataManager trainManager = _tagger
				.resolveDataManagerByArgs();
		SequenceInstance[] trainingData = genTrainingData(_trainFileName,
				trainManager);
		SequenceDataManager evalManager = _tagger
				.resolveDataManagerByArgs();
		SequenceInstance[] evalData = genEvalData(_testFileName,
				_modelFileName, evalManager);

		for (int numIters = 1; numIters <= maxNumOfTrainingIters; numIters++) {
			trainOnData(_trainFileName, _modelFileName, numIters, bAvgParams,
					trainManager, trainingData);
			SequencePredictor predictor = _tagger
					.resolvePredictorByArgs(0);
			predictor.loadModel(_modelFileName);
			if (predictByCommittee != null) {
				predictor = new SequencePredictorByMajorityVote(predictor,
						predictByCommittee);
			}
			evalOnData(predictor, _testFileName, _modelFileName,
					evalResults[numIters - 1], posNegResFile, appendToResFile,
					evalManager, evalData);
		}
		SequenceEvaluationResult.printResultsArray(evalResults);
	}
}
