package struct.alg;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import struct.sequence.SequenceFeatures;
import struct.sequence.SequenceInstance;
import struct.sequence.SequenceLabel;
import struct.solver.QPConstraints;
import struct.solver.QPOutput;
import struct.solver.QPParameters;
import struct.solver.QPSolver;
import struct.types.Alphabet;
import struct.types.FeatureVector;
import struct.types.Features;
import struct.types.Instance;

public class QPBatchUpdators implements BatchUpdator {

    // Parameters + Constructor
    static public enum UpdateType {MMMN_OPTIMAL, MMMN_OPTIMAL_PER_EXAMPLE};
    
    private UpdateType update_type;
    private double c;
    private double message_weight;
    private Alphabet tagAlphabet;
    private int numLabels;
    private int numIters;
    
    private QPSolver qp_solver;

    public QPBatchUpdators(String params, Alphabet ta) {
	update_type = UpdateType.MMMN_OPTIMAL;
	c = Double.POSITIVE_INFINITY;
	QPParameters qp_params = new QPParameters();
	qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B00_FULL;
	numIters = 50;
	
	tagAlphabet = ta;
	numLabels = tagAlphabet.size();
	
	// Decode Params
	System.out.println(params);
	String[] params_str = params.split(" ");
	
	System.out.println("Num of Params : " + params_str.length);
	
	for (int i=0; i<params_str.length; i++) {
	    params_str[i].trim();
	    String[] param = params_str[i].split(":");
	    
	    switch (param[0].charAt(0)) {
		
	    case 'u':
		if (param[1].compareTo("MMMN_OPTIMAL") == 0)
		    update_type = UpdateType.MMMN_OPTIMAL;
		else if (param[1].compareTo("MMMN_OPTIMAL_PER_EXAMPLE") == 0)
		    update_type = UpdateType.MMMN_OPTIMAL_PER_EXAMPLE;
		else
		    throw new IllegalArgumentException("No such update type " + param[1]);
		break;
		
	    case 'C':
		c= Double.parseDouble(param[1].trim());
		break;
		
	    case 'n':
		qp_params.nu = Double.parseDouble(param[1].trim());
		break;


	    case 'm':
		message_weight = Double.parseDouble(param[1].trim());
		break;

	    case 'i':
		qp_params.nu = Integer.parseInt(param[1].trim());
		break;


	    case 's':
		qp_params.eps = Double.parseDouble(param[1].trim());
		break;
		
	    default:
		System.out.println("Unkown option : " + params_str[i]);
	    } // switch
	} // for
	
	
	
	qp_solver = new QPSolver(qp_params);


	System.out.println(this.toString());
	System.out.println(qp_params.toString());
    }


    public String toString() {
	StringBuffer str = new StringBuffer();
	str.append("Update Type: " + update_type + " \n");
	str.append("C          : " + String.valueOf(c) + "\n");
	str.append("message_weight : " + String.valueOf(message_weight) + "\n");
	str.append("numIters   : " + String.valueOf(numIters) + "\n");
	str.append("numLabels  : " + String.valueOf(numLabels) + "\n");
	return (str.toString());
    }


    public void set_c(double c1) {
	c=c1;
    }
        
    public UpdateType get_update_type() {
	return update_type;
    }

    public double get_c() {
	return c;
    }

    

    public void update(Instance[] inst, Predictor predictor, String featureFile) throws IOException {
	long start, end;

	start = System.currentTimeMillis();
	switch (update_type) {
	case MMMN_OPTIMAL:
	    mmmnOptimalUpdate(inst, predictor, featureFile);	    break;
	case MMMN_OPTIMAL_PER_EXAMPLE:
	    mmmnOptimalPerExampleUpdate(inst, predictor, featureFile);	    break;
	default:
	    System.out.println("No such update type " + update_type);
	}
	end = System.currentTimeMillis();

	System.out.println("Time to compute update = " + (end-start));

	
    }



    protected void mmmnOptimalUpdate(Instance[] instances, Predictor predictor, String featureFile) throws IOException {

	
	int n_constraints = 0;
	int n_messages = 0;
	Instance inst;
	Features feats;
	SequenceFeatures sequence_feats;
	SequenceInstance sequence_inst;
	SequenceLabel sequence_label;
	String[] actual_tags;
	int numTokens;
	
	int n_instances = 300;//instances.length;
	System.out.println("n_instances = " + n_instances);
	
	
	// get total number of constraints
	for (int ind=0; ind < n_instances; ind++) {
	    inst = instances[ind];
	    
	    sequence_inst = (SequenceInstance) inst;
	    sequence_label = sequence_inst.getLabel();
	    actual_tags = sequence_label.tags;
	    numTokens = actual_tags.length; 

	    n_constraints += (numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1);
	    n_messages +=  2 * numLabels * numTokens;
	}
	System.out.println("total number of constraints = " + n_constraints);
	System.out.println("total number of messages    = " + n_messages);
	
	n_messages += n_instances; 
	System.out.println("modified total number of messages    = " + n_messages);
	
	// constraints structure
	QPConstraints constraints = new QPConstraints();
	constraints.a1 = new FeatureVector[n_constraints];
	constraints.a2 = new FeatureVector[n_constraints];
	constraints.b  = new double[n_constraints];
	
	constraints.dim1 = predictor.weights.length;
	constraints.dim2 = n_messages;
	constraints.C = c;

	// generate constraints;
	int alpha_base_index;// = 0;
	int beta_base_index;//  = alpha_base_index + numTokens * numLabels + 1;
	int global_base_index=0;

	ObjectInputStream in = new ObjectInputStream(new FileInputStream(featureFile));
	
	int i, k, k1, r;
	int act_ind, act_ind1;

	r = 0;
	for (int ind=0; ind < n_instances; ind++) {

	    inst  = instances[ind];
	    feats = instances[ind].getFeatures(in);
	    
	    sequence_inst = (SequenceInstance) inst;
	    sequence_label = sequence_inst.getLabel();
	    actual_tags = sequence_label.tags;
	    numTokens = actual_tags.length; // length of sequence
	    sequence_feats = (SequenceFeatures) feats;
	    
	    alpha_base_index = global_base_index;	    
	    beta_base_index  = alpha_base_index +  numTokens * numLabels + 1;

	    System.out.println(" ind = " + ind + 
			       " alpha_base_index " + alpha_base_index + 
			       " beta_base_index " +  beta_base_index + 
			       " global_base_index = " + global_base_index);

	
	    // first constraints
	    for (i=1; i<numTokens; i++) {
		act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
		for (k=0; k<numLabels-1; k++) {
		    
		    if (k == act_ind) continue;
		    
		    constraints.a1[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
								      sequence_feats.getFeatureVector(i, k) );
		    
		    constraints.b[r]  = 1;

		    constraints.a2[r] = new FeatureVector(-1 , -1.0 , null);
		    
		    if (i > 0)                                  // forwoard only for i = 1 ... numTokens-1
			constraints.a2[r].add( i * numLabels + k + alpha_base_index, message_weight); // alpha_i[k<act_ind]
		    
		    if (i < numTokens - 1)                      // backwards only for i = 0 ... numTokens-2
			constraints.a2[r].add( i * numLabels + k  + beta_base_index,  message_weight); // -beta_i[k]
		    
		    
		    r++;
		} // for k
	    } // for i
	    
	    // second constraints (going over all edges)
	    
	    for (i=1; i<numTokens-1; i++) {
		act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
		act_ind1 = tagAlphabet.lookupIndex(actual_tags[i+1]);
		
		for (k=0; k<numLabels-1; k++) { // k[i]
		    for (k1=0; k1<numLabels-1; k1++) { // k[i+1]
			if ( (k == act_ind) & (k1 == act_ind1) ) continue;
			
			constraints.a1[r] = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind, act_ind1) , 
									 sequence_feats.getFeatureVector(i, k, k1) );
			
			constraints.b[r] = 0;
			
			constraints.a2[r] = new FeatureVector(-1 , -1.0 , null);
			constraints.a2[r].add( (i+1)  * numLabels + k1  + alpha_base_index, -message_weight); // +alpha_i[k]  
			constraints.a2[r].add( i  * numLabels + k  + beta_base_index,       -message_weight); // +beta_{i+1}[k1]
			
			r++;
		    } // k1
		} //k
		
	    } // for i
	    
	    global_base_index += 2 *  numLabels * numTokens + 1;
	}
	System.out.println("r/n_cosntraints = " + r + "/" + n_constraints);
	
	in.close();
	
	long start, end;
	System.out.println("Running QP");
	start = System.currentTimeMillis();
	QPOutput qp_output;
	qp_output = qp_solver.solve_full(constraints);
	end = System.currentTimeMillis();
	System.out.println("Time run QP: " + (end-start));
	
	System.out.println("predictor.weights.length=" + predictor.weights.length + 
			   " qp_output.w1.lenth=" + qp_output.w1.length);
	for (i=0; i< predictor.weights.length; i++)
	    predictor.weights[i] = qp_output.w1[i];


    }



    protected void mmmnOptimalPerExampleUpdate(Instance[] instances, Predictor predictor, String featureFile) throws IOException {

	double[][] alpha;
	double[]   w1;
	double[][] w2;

	Instance inst;
	Features feats;
	SequenceFeatures sequence_feats;
	SequenceInstance sequence_inst;
	SequenceLabel sequence_label;
	String[] actual_tags;
	int numTokens;
	
	long start, end;
	
	ObjectInputStream in;
	
	System.out.println("allocating memory ... ");

	alpha = new double[instances.length][];
	w2    = new double[instances.length][];
	w1    = new double[predictor.weights.length];
	
	start = System.currentTimeMillis();

	in = new ObjectInputStream(new FileInputStream(featureFile));

	for (int j = 0; j < instances.length; j++) {
	    
	    inst = instances[j];
	    feats = instances[j].getFeatures(in);
	    sequence_inst = (SequenceInstance) inst;
	    sequence_label = sequence_inst.getLabel();
	    actual_tags = sequence_label.tags;
	    numTokens = actual_tags.length; 
		
	    w2[j]    = new double[2*numLabels*numTokens];
	    alpha[j] = new double[(numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1)];
	}

	in.close();


	end = System.currentTimeMillis();
	System.out.println("Time to allocating memory " + (end-start));


	// training
	
	QPConstraints constraints;
	QPOutput qp_output;
	for (int i = 0; i < numIters; i++) {
	    
	    System.out.println("==========================");
	    System.out.println("Training iteration: " + i);
	    System.out.println("==========================");
	    
	    long start_iter = System.currentTimeMillis();

	    in = new ObjectInputStream(new FileInputStream(featureFile));

	    for (int j = 0; j < instances.length; j++) {
		
		inst = instances[j];
		feats = instances[j].getFeatures(in);
		
		start = System.currentTimeMillis();
		constraints = generateConstraintsForMMMNOptimalPerExample(inst, feats, w1.length);
		System.out.println("Time to generate constraints: " + (end-start));

		constraints.w1 = w1;
		constraints.w2 = w2[j];
		constraints.alpha = alpha[j];
		start = System.currentTimeMillis();
		qp_output = qp_solver.solve_full(constraints);
		end = System.currentTimeMillis();
		System.out.println("Time run QP: " + (end-start));

		w1       = qp_output.w1;
		w2[j]    = qp_output.w2;
		alpha[j] = qp_output.alpha;
	
	    }
	    
	    in.close();
	    
	    long end_iter = System.currentTimeMillis();
	    System.out.println("Training took: " + (end_iter-start_iter));

	    

	}
	System.out.println("");


	for (int ii=0; ii< predictor.weights.length; ii++)
	    predictor.weights[ii] = w1[ii];


    }



    protected QPConstraints generateConstraintsForMMMNOptimalPerExample(Instance inst, Features feats, int dim1) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;

	int alpha_base_index = 0;
	int beta_base_index  = alpha_base_index + numTokens * numLabels;
	System.out.println("Alpha Base Index of Param " + alpha_base_index);
	System.out.println("Beta  Base Index of Param " + beta_base_index);
	int n_cosntraints = (numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1);
	System.out.println("Number of Constraints (n_cosntraints) = " + n_cosntraints);

	QPConstraints constraints = new QPConstraints();

	constraints.a1 = new FeatureVector[n_cosntraints];
	constraints.a2 = new FeatureVector[n_cosntraints];
	constraints.a  = constraints.a1;
	constraints.b  = new double[n_cosntraints];
	
	constraints.dim1 = dim1;
	constraints.dim2 = 2 * numLabels * numTokens ;

	constraints.C = c;
	
	int i, k, k1, r;
	int act_ind, act_ind1;

	
	r = 0;
	
	// first constraints
	for (i=1; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels-1; k++) {
		
		if (k == act_ind) continue;
		
		constraints.a1[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
								  sequence_feats.getFeatureVector(i, k) );

		constraints.b[r]  = 1;
		
		constraints.a2[r] = new FeatureVector(-1 , -1.0 , null);
		
		if (i > 0)                                  // forwoard only for i = 1 ... numTokens-1
		    constraints.a2[r].add( i * numLabels + k + alpha_base_index, message_weight); // alpha_i[k<act_ind]
		
		if (i < numTokens - 1)                      // backwards only for i = 0 ... numTokens-2
		    constraints.a2[r].add( i * numLabels + k  + beta_base_index,  message_weight); // -beta_i[k]
		
		
		
		r++;
	    } // for k
	} // for i
	
	// second constraints (going over all edges)
	FeatureVector temp1, temp2, temp3, temp4;
	for (i=1; i<numTokens-1; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    act_ind1 = tagAlphabet.lookupIndex(actual_tags[i+1]);
	    
	    for (k=0; k<numLabels-1; k++) { // k[i]
		for (k1=0; k1<numLabels-1; k1++) { // k[i+1]
		    if ( (k == act_ind) & (k1 == act_ind1) ) continue;
		    
		    constraints.a1[r] = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind, act_ind1) , 
								     sequence_feats.getFeatureVector(i, k, k1) );

		    constraints.b[r] = 0;

		    constraints.a2[r] = new FeatureVector(-1 , -1.0 , null);
		    constraints.a2[r].add( (i+1)  * numLabels + k1  + alpha_base_index, -message_weight); // +alpha_i[k]  
		    constraints.a2[r].add( i  * numLabels + k  + beta_base_index, -message_weight); // +beta_{i+1}[k1]
		    
		    r++;
		} // k1
	    } //k
	    
	} // for i
	
	System.out.println("r/n_cosntraints = " + r + "/" + n_cosntraints);
        
    
    

	return constraints;
    }




}


