package struct.sequence;

import cw_library.tools.MathTools;

public class ConfEstimation {
	public enum WordConfEvalType {
		committee, weightedCommittee
	}

	public enum SeqConfEvalType {
		minOfWords, avgOfWords
	}

	static public SeqConfEvalType resolveFromString(String str) {
		if (str.equalsIgnoreCase("min")) {
			return SeqConfEvalType.minOfWords;
		} else if (str.equalsIgnoreCase("avg")) {
			return SeqConfEvalType.avgOfWords;
		} else {
			throw new RuntimeException("Unknown Type: " + str);
		}

	}

	//
	// Committee is a set of K predictions of the same sentence (such as K-best
	// Viterbi or K-Draw)
	// The 'chosen' tag per word is the tag specified by label#0.
	// The confidence is the percent of agreement on the chosen tag among the K
	// predictions.
	// Can be [1/K,1]
	//
	static public double evalWordConfByCommittee(
			SequencePrediction seqPrediction, int k, int wordIdx) {
		SequenceLabel[] labels = seqPrediction.getAllLabels();
		SequenceLabel bestLabel = labels[0];
		String chosenTag = bestLabel.tags[wordIdx];

		int voteCount = 0;
		for (int i = 0; i < k; i++) {
			if (labels[i] == null) {
				// May happen if specified K is greater that actual available
				// predictions
				break;
			}

			String tag = labels[i].tags[wordIdx];
			if (tag.equals(chosenTag)) {
				voteCount++;
			}
		}

		double conf = (double) voteCount / (double) k;
		return conf;
	}

	//
	// Committee is a set of K predictions of the same sentence (such as K-best
	// Viterbi or K-Draw)
	// The 'chosen' tag per word is the tag specified by label#0.
	// The confidence is the weighted votes regarding the chosen tag among the K
	// predictions, weighted by the prediction's probability.
	//
	static public double evalWordConfByWeightedCommittee(
			SequencePrediction seqPrediction, int k, int wordIdx) {
		SequenceLabel[] labels = seqPrediction.getAllLabels();
		SequenceLabel bestLabel = labels[0];
		String chosenTag = bestLabel.tags[wordIdx];

		double totalProb = 0;
		double agreementProb = 0;
		for (int i = 0; i < k; i++) {
			if (labels[i] == null) {
				// May happen if specified K is greater that actual available
				// predictions
				break;
			}
			String tag = labels[i].tags[wordIdx];
			double prob = labels[i].prob;
			// Weighted mean is defined only for non-negative weights
			if (prob > 0) {
				totalProb += prob;
				if (tag.equals(chosenTag)) {
					agreementProb += prob;
				}
			} else {
				// rest of labels will have even lower score...
				break;
			}
		}

		double conf = agreementProb / totalProb;
		return conf;
	}

	static public double evalWordConfBy(WordConfEvalType wordConfEvalType,
			SequencePrediction seqPrediction, int k, int wordIdx) {
		switch (wordConfEvalType) {
		case committee:
			return evalWordConfByCommittee(seqPrediction, k, wordIdx);
		case weightedCommittee:
			return evalWordConfByWeightedCommittee(seqPrediction, k, wordIdx);
		}
		throw new RuntimeException("Unknown eval type: " + wordConfEvalType);
	}

	//
	// evaluate entire sequence by one of several possible methods
	//
	static public double evalSeqConfByCommittee(
			SeqConfEvalType seqConfEvalType, WordConfEvalType wordConfEvalType,
			SequencePrediction seqPrediction) {
		return evalSeqConfByCommittee(seqConfEvalType, wordConfEvalType,
				seqPrediction, seqPrediction.labels.length);
	}

	static public double evalSeqConfByCommittee(
			SeqConfEvalType seqConfEvalType, WordConfEvalType wordConfEvalType,
			SequencePrediction seqPrediction, int k) {

		// First evaluate the confidence of each word
		int numOfWords = seqPrediction.getBestLabel().tags.length;
		double[] wordScores = new double[numOfWords];
		for (int wordIdx = 0; wordIdx < numOfWords; wordIdx++) {
			wordScores[wordIdx] = evalWordConfBy(wordConfEvalType,
					seqPrediction, k, wordIdx);
		}

		//
		// Now evaluate the confidence of the entire sequence by one of the
		// possible methods
		//
		switch (seqConfEvalType) {
		case minOfWords:
			return MathTools.minValue(wordScores);
		case avgOfWords:
			return MathTools.calculateMean(wordScores);
		}
		return 0;
	}

	static public double evalSeqConfBySumOfCommittees(
			SeqConfEvalType seqConfEvalType, WordConfEvalType wordConfEvalType,
			SequencePrediction[] seqPredictions) {

		// First evaluate the confidence of each word
		int numOfWords = seqPredictions[0].getBestLabel().tags.length;
		double[] wordScores = new double[numOfWords];
		for (int predIdx = 0; predIdx < seqPredictions.length; predIdx++) {
			SequencePrediction seqPrediction = seqPredictions[predIdx];
			int k = seqPrediction.labels.length;
			for (int wordIdx = 0; wordIdx < numOfWords; wordIdx++) {
				wordScores[wordIdx] += evalWordConfBy(wordConfEvalType,
						seqPrediction, k, wordIdx);
			}
		}

		//
		// Now evaluate the confidence of the entire sequence by one of the
		// possible methods
		//
		switch (seqConfEvalType) {
		case minOfWords:
			return MathTools.minValue(wordScores);
		case avgOfWords:
			return MathTools.calculateMean(wordScores);
		}
		return 0;
	}
}
