package struct.sequence;

import struct.types.Alphabet;
import struct.types.FeatureVector;
import struct.types.Features;
import struct.types.Instance;

public class ConfidenceEstimatorDelta extends ConfidenceEstimator {

	public void estimateInst(Instance inst, Features feats,
			LabelConfScore[] allWordsScores, SequencePredictor predictor) {
		genDeltaTable(inst, feats, predictor);

		for (int wordIdx = 0; wordIdx < allWordsScores.length; wordIdx++) {
			double score = getDeltaDiff(wordIdx);
			allWordsScores[wordIdx].score = score;
		}
	}

	public void estimateInst(Instance inst, Features feats,
			double[] allWordsScores, SequencePredictor predictor) {
		genDeltaTable(inst, feats, predictor);

		for (int wordIdx = 0; wordIdx < allWordsScores.length; wordIdx++) {
			double score = getDeltaDiff(wordIdx);
			allWordsScores[wordIdx] = score;
		}
	}

	//
	// Table that holds for every possible constraint [token][tag] the best
	// viterbi score that complies with this constraint
	//
	double bestDecodeScoreNoConstraint;
	double[] minConstrainedScoreDiffFromBest;

	double getDeltaDiff(int wordIdx) {
		return minConstrainedScoreDiffFromBest[wordIdx];
	}

	double[][] cacheTokenTagProb;
	double[][][] cacheTokenPrevTagTagProb;

	void genDeltaTable(Instance inst, Features feats,
			SequencePredictor predictor) {

		SequenceInstance sinst = (SequenceInstance) inst;
		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();
		String[] toks = sinst.getInput().sentence;
		double[][] constrainedScore = new double[toks.length][tagAlphabet
				.size()];
		cacheTokenTagProb = new double[toks.length][tagAlphabet.size()];
		cacheTokenPrevTagTagProb = new double[toks.length][tagAlphabet.size()][tagAlphabet
				.size()];
		calcCacheProbs(inst, feats, predictor);

		//
		// Calc the best score with no constraint
		//
		bestDecodeScoreNoConstraint = decodeNoConstraints(inst, feats,
				predictor);

		//
		// Now calc scores with all possible constraints
		//
		for (int tokIdx = 0; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				constrainedScore[tokIdx][tagIdx] = constrainedDecode(inst,
						feats, tokIdx, tagIdx, predictor);
			}
		}

		//
		// Now calc for each token the minimal difference between the true best
		// score to the best constrained score for this token (which is the
		// second best score)
		//
		minConstrainedScoreDiffFromBest = new double[toks.length];
		for (int tokIdx = 0; tokIdx < toks.length; tokIdx++) {
			double bestScore = Double.NEGATIVE_INFINITY;
			double secondBestScore = Double.NEGATIVE_INFINITY;
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				if (bestScore <= constrainedScore[tokIdx][tagIdx]) {
					secondBestScore = bestScore;
					bestScore = constrainedScore[tokIdx][tagIdx];
				} else if (secondBestScore <= constrainedScore[tokIdx][tagIdx]) {
					secondBestScore = constrainedScore[tokIdx][tagIdx];
				}
			}

			//
			// now calc the diff between second best and true (non-constrained)
			// best
			//
			double diff = bestDecodeScoreNoConstraint - secondBestScore;
			minConstrainedScoreDiffFromBest[tokIdx] = diff;
		}

	}

	private void calcCacheProbs(Instance inst, Features feats,
			SequencePredictor predictor) {
		double[] weightsVector = predictor.weights;
		SequenceInstance sinst = (SequenceInstance) inst;
		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();
		String[] toks = sinst.getInput().sentence;
		SequenceFeatures sf = (SequenceFeatures) feats;

		// Start with calculating for the first token
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			FeatureVector fv_i = sf.getFeatureVector(0, tagIdx);
			double prob_i = predictor.score(fv_i, weightsVector);
			cacheTokenTagProb[0][tagIdx] = prob_i;
		}

		// Now go over all Tokens from first to last, and all possible Tags
		for (int tokIdx = 1; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				FeatureVector fv_i = sf.getFeatureVector(tokIdx, tagIdx);
				double prob_i = predictor.score(fv_i, weightsVector);
				cacheTokenTagProb[tokIdx][tagIdx] = prob_i;
				for (int prevTagIdx = 0; prevTagIdx < tagAlphabet.size(); prevTagIdx++) {
					// Calculate the score according to features of
					// Token-Tag-PrevTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx,
							prevTagIdx, tagIdx);
					double prob_ij = predictor.score(fv_ij, weightsVector);
					cacheTokenPrevTagTagProb[tokIdx][prevTagIdx][tagIdx] = prob_ij;
				}
			}
		}
	}

	double decodeNoConstraints(Instance inst, Features feats,
			SequencePredictor predictor) {
		// use invalid token and tag indices so the constraint will have no
		// effect
		return constrainedDecode(inst, feats, -1, -1, predictor);
	}

	double constrainedDecode(Instance inst, Features feats, int forceTokenIdx,
			int forceTagIdx, SequencePredictor predictor) {
		// double[] weightsVector = predictor.weights;
		SequenceInstance sinst = (SequenceInstance) inst;

		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();

		String[] toks = sinst.getInput().sentence;

		// This is the 2D delta table of [Tokens][Tags] - that holds the best
		// score of a sequences that end at each [Token][Tag].
		double[][] delta = new double[toks.length][tagAlphabet.size()];

		// SequenceFeatures sf = (SequenceFeatures) feats;

		// Calculate delta
		// Start with calculating for the first token
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			/*
			 * FeatureVector fv_i = sf.getFeatureVector(0, tagIdx); double
			 * prob_i = predictor.score(fv_i, weightsVector);
			 */
			double prob_i = cacheTokenTagProb[0][tagIdx];
			delta[0][tagIdx] = prob_i;
		}
		// Apply constraints
		if (0 == forceTokenIdx) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				if (tagIdx == forceTagIdx) {
					continue;
				}
				// else - force -infinity
				delta[0][tagIdx] = Double.NEGATIVE_INFINITY;
			}
		}

		// Now go over all Tokens from first to last, and all possible Tags
		for (int tokIdx = 1; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				/*
				 * FeatureVector fv_i = sf.getFeatureVector(tokIdx, tagIdx);
				 * double prob_i = predictor.score(fv_i, weightsVector);
				 */
				double prob_i = cacheTokenTagProb[tokIdx][tagIdx];

				// For each Token-tag slot check all possible previous Tags
				double bestPrev = Double.NEGATIVE_INFINITY;
				for (int prevTagIdx = 0; prevTagIdx < tagAlphabet.size(); prevTagIdx++) {
					// Calculate the score according to features of
					// Token-Tag-PrevTag
					/*
					 * FeatureVector fv_ij = sf.getFeatureVector(tokIdx,
					 * prevTagIdx, tagIdx); double prob_ij =
					 * predictor.score(fv_ij, weightsVector);
					 */
					double prob_ij = cacheTokenPrevTagTagProb[tokIdx][prevTagIdx][tagIdx];
					double delta_ij = prob_ij + delta[tokIdx - 1][prevTagIdx];

					if (bestPrev < delta_ij) {
						bestPrev = delta_ij;
					}
				}

				delta[tokIdx][tagIdx] = bestPrev + prob_i;
			}

			// Apply constraints
			if (tokIdx == forceTokenIdx) {
				for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
					if (tagIdx == forceTagIdx) {
						continue;
					}
					// else - force -infinity
					delta[tokIdx][tagIdx] = Double.NEGATIVE_INFINITY;
				}
			}
		}

		//
		// Now find the best score that complies with the constraints
		//
		int lastToken = toks.length - 1;
		double bestScore = Double.NEGATIVE_INFINITY;
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			if (bestScore < delta[lastToken][tagIdx]) {
				bestScore = delta[lastToken][tagIdx];
			}
		}
		return bestScore;
	}
}
