package struct.sequence;

import struct.types.Alphabet;
import struct.types.FeatureVector;
import struct.types.Features;
import struct.types.Instance;

public class ConfidenceEstimatorGammaExp extends ConfidenceEstimator {

	double coef;

	public ConfidenceEstimatorGammaExp(double coef) {
		// the coefficient used for Exp[coef*<w*phi(pos,tag)>]
		this.coef = coef;
	}

	public void estimateInst(Instance inst, Features feats,
			LabelConfScore[] allWordsScores, SequencePredictor predictor) {
		// genGammaTable(inst, feats, predictor);
		genGammaTable_logs(inst, feats, predictor);

		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();
		SequencePrediction prediction = predictor.decode(inst, feats);
		String[] predictedTags = ((SequenceLabel) prediction.getBestLabel()).tags;

		for (int wordIdx = 0; wordIdx < allWordsScores.length; wordIdx++) {
			int tagIdx = tagAlphabet.lookupIndex(predictedTags[wordIdx]);
			allWordsScores[wordIdx].score = gamma[wordIdx][tagIdx];
		}
	}

	//
	// Gamma data tables
	//
	double[][] gamma;

	public void genGammaTable(Instance inst, Features feats,
			SequencePredictor predictor) {
		double probDiv = 1E40;
		double[] weightsVector = predictor.weights;
		SequenceInstance sinst = (SequenceInstance) inst;

		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();

		String[] toks = sinst.getInput().sentence;

		// This is the 2D alpha table of [Tokens][Tags] - that holds the total
		// score of a sequences that end at each [Token][Tag].
		double[][] alpha = new double[toks.length][tagAlphabet.size()];

		// And beta table that holds the score of a sequences that start at each
		// [Token][Tag].
		double[][] beta = new double[toks.length][tagAlphabet.size()];

		double probNormalizer = 0;

		SequenceFeatures sf = (SequenceFeatures) feats;

		// Calculate alpha
		// Start with calculating for the first token
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			FeatureVector fv_i = sf.getFeatureVector(0, tagIdx);
			double score_i = predictor.score(fv_i, weightsVector);
			score_i *= coef;
			double prob_i = Math.exp(score_i);
			// alpha(0, tag) = exp[w*phi(0,tag)]
			alpha[0][tagIdx] = prob_i;
			alpha[0][tagIdx] /= probDiv;
		}

		// Now go over all Tokens from first to last, and all possible Tags
		for (int tokIdx = 1; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				FeatureVector fv_i = sf.getFeatureVector(tokIdx, tagIdx);
				double score_i = predictor.score(fv_i, weightsVector);
				score_i *= coef;
				// b(pos, tag) = exp[w*phi(pos,tag)]
				double prob_i = Math.exp(score_i);

				// For each Token-tag slot check all possible previous Tags
				double totTransitionProb = 0;
				for (int prevTagIdx = 0; prevTagIdx < tagAlphabet.size(); prevTagIdx++) {
					// Calculate the score according to features of
					// Token-Tag-PrevTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx,
							prevTagIdx, tagIdx);
					double score_ij = predictor.score(fv_ij, weightsVector);
					score_ij *= coef;
					// A(pos,tag_i,tag_j) = exp[w*phi(pos,tag_i,tag_j)]
					double prob_ij = Math.exp(score_ij);

					// Get the prevToken-prevTag item
					// Sum over all i : (alpha(pos-1,tag_i)*A(pos,tag_i,tag_j)
					totTransitionProb += (prob_ij * alpha[tokIdx - 1][prevTagIdx]);
				}

				// alpha(j, tag) = (sum of transition from all possible prev
				// tags) * b(pos, tag)
				alpha[tokIdx][tagIdx] = totTransitionProb * prob_i;
			}
		}

		// Calculate beta
		int lastTokenIdx = toks.length - 1;
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			beta[lastTokenIdx][tagIdx] = 1;
		}

		// Now go over all Tokens from last-1 to first, and all possible Tags
		for (int tokIdx = lastTokenIdx - 1; tokIdx >= 0; tokIdx--) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				double totProb = 0;

				// For each Token-tag slot check all possible next Tags
				for (int nextTagIdx = 0; nextTagIdx < tagAlphabet.size(); nextTagIdx++) {
					// Calculate the score according to features of
					// Token-NextTag
					FeatureVector fv_j = sf.getFeatureVector(tokIdx + 1,
							nextTagIdx);
					double score_j = predictor.score(fv_j, weightsVector);
					score_j *= coef;
					double prob_j = Math.exp(score_j);
					// Calculate the score according to features of
					// Token-Tag-NextTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx + 1,
							tagIdx, nextTagIdx);
					double score_ij = predictor.score(fv_ij, weightsVector);
					score_ij *= coef;
					double prob_ij = Math.exp(score_ij);

					// Get the nextToken-nextTag
					totProb += (prob_ij * beta[tokIdx + 1][nextTagIdx] * prob_j);
				}

				beta[tokIdx][tagIdx] = totProb;
			}
		}

		// Calculate normalizer as sum of all alphas at last position
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			probNormalizer += alpha[lastTokenIdx][tagIdx];
		}

		// Calc Gamma table
		gamma = new double[toks.length][tagAlphabet.size()];
		for (int tokIdx = 0; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				gamma[tokIdx][tagIdx] = alpha[tokIdx][tagIdx]
						* beta[tokIdx][tagIdx] / probNormalizer;
			}
		}

		return;
	}

	double sumOfLogs(double a, double b) {
		if (a == Double.NEGATIVE_INFINITY) {
			if (b == Double.NEGATIVE_INFINITY)
				return Double.NEGATIVE_INFINITY;
			return b;
		} else if (b == Double.NEGATIVE_INFINITY)
			return a;
		else if (a > b)
			return a + Math.log(1 + Math.exp(b - a));
		else
			return b + Math.log(1 + Math.exp(a - b));
	}

	public void genGammaTable_logs(Instance inst, Features feats,
			SequencePredictor predictor) {
		double[] weightsVector = predictor.weights;
		SequenceInstance sinst = (SequenceInstance) inst;

		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();

		String[] toks = sinst.getInput().sentence;

		// This is the 2D alpha table of [Tokens][Tags] - that holds the total
		// score of a sequences that end at each [Token][Tag].
		double[][] alpha = new double[toks.length][tagAlphabet.size()];

		// And beta table that holds the score of a sequences that start at each
		// [Token][Tag].
		double[][] beta = new double[toks.length][tagAlphabet.size()];

		SequenceFeatures sf = (SequenceFeatures) feats;

		// Calculate alpha
		// Start with calculating for the first token
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			FeatureVector fv_i = sf.getFeatureVector(0, tagIdx);
			double score_i = predictor.score(fv_i, weightsVector);
			score_i *= coef;
			double prob_i = score_i;
			// alpha(0, tag) = [w*phi(0,tag)]
			alpha[0][tagIdx] = prob_i;
		}

		// Now go over all Tokens from first to last, and all possible Tags
		for (int tokIdx = 1; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				FeatureVector fv_i = sf.getFeatureVector(tokIdx, tagIdx);
				double score_i = predictor.score(fv_i, weightsVector);
				score_i *= coef;
				// b(pos, tag) = [w*phi(pos,tag)]
				double prob_i = score_i;

				// For each Token-tag slot check all possible previous Tags
				double totTransitionProb = Double.NEGATIVE_INFINITY;
				for (int prevTagIdx = 0; prevTagIdx < tagAlphabet.size(); prevTagIdx++) {
					// Calculate the score according to features of
					// Token-Tag-PrevTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx,
							prevTagIdx, tagIdx);
					double score_ij = predictor.score(fv_ij, weightsVector);
					score_ij *= coef;
					// A(pos,tag_i,tag_j) = [w*phi(pos,tag_i,tag_j)]
					double prob_ij = score_ij;

					// Get the prevToken-prevTag item
					// Sum over all i : (alpha(pos-1,tag_i)*A(pos,tag_i,tag_j)
					totTransitionProb = sumOfLogs(totTransitionProb,
							(prob_ij + alpha[tokIdx - 1][prevTagIdx]));
					// totTransitionProb += Math.exp(prob_ij + alpha[tokIdx -
					// 1][prevTagIdx]);
				}

				// alpha(j, tag) = (sum of transition from all possible prev
				// tags) * b(pos, tag)
				// alpha[tokIdx][tagIdx] = Math.log(totTransitionProb) + prob_i;
				alpha[tokIdx][tagIdx] = totTransitionProb + prob_i;
			}
		}

		// Calculate beta
		int lastTokenIdx = toks.length - 1;
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			beta[lastTokenIdx][tagIdx] = 0; // Log(1) --> 0
		}

		// Now go over all Tokens from last-1 to first, and all possible Tags
		for (int tokIdx = lastTokenIdx - 1; tokIdx >= 0; tokIdx--) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				double totProb = Double.NEGATIVE_INFINITY;

				// For each Token-tag slot check all possible next Tags
				for (int nextTagIdx = 0; nextTagIdx < tagAlphabet.size(); nextTagIdx++) {
					// Calculate the score according to features of
					// Token-NextTag
					FeatureVector fv_j = sf.getFeatureVector(tokIdx + 1,
							nextTagIdx);
					double score_j = predictor.score(fv_j, weightsVector);
					score_j *= coef;
					double prob_j = score_j;
					// Calculate the score according to features of
					// Token-Tag-NextTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx + 1,
							tagIdx, nextTagIdx);
					double score_ij = predictor.score(fv_ij, weightsVector);
					score_ij *= coef;
					double prob_ij = score_ij;

					// Get the nextToken-nextTag
					// totProb += Math.exp(prob_ij + beta[tokIdx +
					// 1][nextTagIdx] + prob_j);
					totProb = sumOfLogs(totProb, (prob_ij
							+ beta[tokIdx + 1][nextTagIdx] + prob_j));
				}

				// beta[tokIdx][tagIdx] = Math.log(totProb);
				beta[tokIdx][tagIdx] = totProb;
			}
		}

		// Calculate normalizer as sum of all alphas at last position
		// double probNormalizer = 0;
		double probNormalizer = Double.NEGATIVE_INFINITY;
		for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
			// probNormalizer += Math.exp(alpha[lastTokenIdx][tagIdx]);
			probNormalizer = sumOfLogs(probNormalizer,
					alpha[lastTokenIdx][tagIdx]);
		}

		// Calc Gamma table
		gamma = new double[toks.length][tagAlphabet.size()];
		for (int tokIdx = 0; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {
				gamma[tokIdx][tagIdx] = Math.exp(alpha[tokIdx][tagIdx]
						+ beta[tokIdx][tagIdx] - probNormalizer);
			}
		}

		return;
	}
}
