package struct.sequence;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Random;

import struct.alg.Predictor;
import struct.types.Alphabet;
import struct.types.FeatureVector;
import struct.types.Features;
import struct.types.Instance;

public class SequencePredictor extends Predictor {

	public double[] sigma;

	public SequencePredictor(int dimensions) {
		super(dimensions);
		setCommitteeType(PredictCommitteeType.kBestViterbi);
	}

	public void saveModel(String file) throws Exception {
		System.out.print("Saving model ... ");
		ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(
				file));
		out.writeObject(weights);
		out.writeObject(SequenceInstance.getDataAlphabet());
		out.writeObject(SequenceInstance.getTagAlphabet());
		out.close();
		System.out.println("done.");
	}

	public void loadModel(String file) throws Exception {
		System.out.print("Loading model ... ");
		ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
		weights = (double[]) in.readObject();
		SequenceInstance.setDataAlphabet((Alphabet) in.readObject());
		SequenceInstance.setTagAlphabet((Alphabet) in.readObject());
		in.close();
		System.out.println("done.");
	}

	public SequencePrediction decode(Instance inst, Features feats) {
		return decode(inst, feats, 1);
	}

	// Return K-Best predictions, either by drawing K classifiers or by using
	// K-Best Viterbi
	public SequencePrediction decode(Instance inst, Features feats, int K) {
		if (committeeType == PredictCommitteeType.kBestViterbi) {
			return decode(inst, feats, K, this.weights);
		}

		// else - draw K classifiers
		drawKWeightsVectors(K);

		SequenceLabel[] labels = new SequenceLabel[K];
		for (int i = 0; i < K; i++) {
			SequencePrediction prediction = decode(inst, feats, 1,
					weightVectors[i]);
			labels[i] = prediction.getBestLabel();
		}

		SequencePrediction predictions = new SequencePrediction(labels);
		return predictions;
	}

	// The K-Best Viterbi algorithm
	public SequencePrediction decode(Instance inst, Features feats, int K,
			double[] weightsVector) {

		KBestSequence kBestSeqs = decodeKBestSequences(inst, feats, K,
				weightsVector);

		// Finally, extract from the 3D table of [Tokens][Tags][K] the best K
		// sequences
		return kBestSeqs.getBestSequences();
	}

	public KBestSequence decodeKBestSequences(Instance inst, Features feats,
			int K, double[] weightsVector) {
		SequenceInstance sinst = (SequenceInstance) inst;

		Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();

		String[] toks = sinst.getInput().sentence;

		// This is the 3D table of [Tokens][Tags][K] - that holds the best K
		// sequences that end at each [Token][Tag].
		KBestSequence kBestSeqs = new KBestSequence(tagAlphabet, sinst, K);

		SequenceFeatures sf = (SequenceFeatures) feats;

		kBestSeqs.add(0, tagAlphabet.lookupIndex("O"), 0.0, new FeatureVector(
				-1, -1.0, null), null, 0);
		for (int i = 0; i < tagAlphabet.size(); i++)
			kBestSeqs.add(0, i, Double.NEGATIVE_INFINITY, new FeatureVector(-1,
					-1.0, null), null, 0);

		// Now go over all Tokens from first to last, and all possible Tags
		for (int tokIdx = 1; tokIdx < toks.length; tokIdx++) {
			for (int tagIdx = 0; tagIdx < tagAlphabet.size(); tagIdx++) {

				// For each Token-tag slot check all possible previous Tags
				for (int prevTagIdx = 0; prevTagIdx < tagAlphabet.size(); prevTagIdx++) {

					// Calculate the score according to features of
					// Token-Tag-PrevTag
					FeatureVector fv_ij = sf.getFeatureVector(tokIdx,
							prevTagIdx, tagIdx);
					double prob_ij = score(fv_ij, weightsVector);
					FeatureVector fv_i = sf.getFeatureVector(tokIdx, tagIdx);
					double prob_i = score(fv_i, weightsVector);

					// Get the best K sequences to reach the prevToken-prevTag
					SequenceItem[] kBestPrevItems = kBestSeqs.getItems(
							tokIdx - 1, prevTagIdx);

					if (kBestPrevItems == null)
						continue;

					int startInsertIdx = 0;
					for (int i = 0; i < kBestPrevItems.length; i++) {
						if (kBestPrevItems[i].prob == Double.NEGATIVE_INFINITY) {
							// There was no i'th way to reach the
							// prevToken-prevTag
							continue;
						}
						// The 'startInsertIdx' is just for optimization - tells
						// the index, out of best K, where the last sequence was
						// inserted at. So (1) the next insert, which is known
						// as worst than current will start just there and not
						// scan from the beginning, and (2) when it is returned
						// as '-1', this means the sequence wasn't one of best K
						// at all so we can stop the search
						startInsertIdx = kBestSeqs.add(tokIdx, tagIdx, prob_ij
								+ prob_i + kBestPrevItems[i].prob,
								FeatureVector.cat(fv_ij, fv_i),
								kBestPrevItems[i], startInsertIdx);
						if (startInsertIdx < 0)
							break;
					}

				}

			}
		}

		// Finally, return the K-Best sequences
		return kBestSeqs;
	}

	public class KBestSequence {

		// This is the 3D table of [Tokens][Tags][K] - that holds the best K
		// sequences that end at each [Token][Tag].
		private SequenceItem[][][] chart;
		private int K;
		private int num_tags;
		private Alphabet tagAlphabet;

		public KBestSequence(Alphabet tagAlphabet, SequenceInstance inst, int K) {
			this.K = K;
			chart = new SequenceItem[inst.getInput().sentence.length][tagAlphabet
					.size()][K];
			num_tags = tagAlphabet.size();
			this.tagAlphabet = tagAlphabet;
		}

		// Add to the proper rank among the best-K a new sequence option that
		// ends at Token-Tag with given 'prob'. The FeaturesVector are the
		// features that were used to score according to the Token-Tag-PrevTag.
		// 
		// The 'startInsertIdx' is just for optimization - tells
		// the index, out of best K, where the last sequence was
		// inserted at. So (1) the next insert, if it is known
		// as worst than current will start just there and not
		// scan from the beginning.
		//
		// Return value - the rank where the new sequence was added, '-1' if not
		// added at all since the 'prob' is too low.
		public int add(int tokIdx, int tagIdx, double prob, FeatureVector fv,
				SequenceItem prev, int startInsertIdx) {

			if (chart[tokIdx][tagIdx][0] == null) {
				for (int i = 0; i < K; i++)
					chart[tokIdx][tagIdx][i] = new SequenceItem(tokIdx, tagIdx,
							Double.NEGATIVE_INFINITY, null, null);
			}

			if (chart[tokIdx][tagIdx][K - 1].prob > prob)
				return -1;

			for (int i = startInsertIdx; i < K; i++) {
				if (chart[tokIdx][tagIdx][i].prob < prob) {
					SequenceItem tmp = chart[tokIdx][tagIdx][i];
					chart[tokIdx][tagIdx][i] = new SequenceItem(tokIdx, tagIdx,
							prob, fv, prev);
					for (int j = i + 1; j < K
							&& tmp.prob != Double.NEGATIVE_INFINITY; j++) {
						SequenceItem tmp1 = chart[tokIdx][tagIdx][j];
						chart[tokIdx][tagIdx][j] = tmp;
						tmp = tmp1;
					}
					return i + 1 >= K ? -1 : i + 1;
				}

			}
			return -1;
		}

		public double getProb(int tokIdx, int tagIdx) {
			return getProb(tokIdx, tagIdx, 0);
		}

		public double getProb(int tokIdx, int tagIdx, int i) {
			if (chart[tokIdx][tagIdx][i] != null)
				return chart[tokIdx][tagIdx][i].prob;
			return Double.NEGATIVE_INFINITY;
		}

		public double[] getProbs(int tokIdx, int tagIdx) {
			double[] result = new double[K];
			for (int i = 0; i < K; i++)
				result[i] = chart[tokIdx][tagIdx][i] != null ? chart[tokIdx][tagIdx][i].prob
						: Double.NEGATIVE_INFINITY;
			return result;
		}

		public SequenceItem getItem(int tokIdx, int tagIdx) {
			return getItem(tokIdx, tagIdx, 0);
		}

		public SequenceItem getItem(int tokIdx, int tagIdx, int i) {
			if (chart[tokIdx][tagIdx][i] != null)
				return chart[tokIdx][tagIdx][i];
			return null;
		}

		public SequenceItem[] getItems(int tokIdx, int tagIdx) {
			if (chart[tokIdx][tagIdx][0] != null)
				return chart[tokIdx][tagIdx];
			else {
				System.out.println(tokIdx + " " + tagIdx);
				System.exit(0);
			}
			return null;
		}

		public SequenceItem[] getBestSequencesAsItems() {

			int lastTokIdx = chart.length - 1;

			SequenceItem[] best = new SequenceItem[K];
			for (int i = 0; i < K; i++) {
				best[i] = new SequenceItem(-1, -1, Double.NEGATIVE_INFINITY,
						null, null);
			}

			for (int tagIdx = 0; tagIdx < num_tags; tagIdx++) {
				for (int k = 0; k < K; k++) {
					// Go over all K x Tags sequences ending at the last token
					// and insert them to the proper rank in the best K
					SequenceItem cand = chart[lastTokIdx][tagIdx][k];

					for (int i = 0; i < K; i++) {
						if (best[i].prob < cand.prob) {
							SequenceItem tmp = best[i];
							best[i] = cand;
							for (int j = i + 1; j < K
									&& tmp.prob != Double.NEGATIVE_INFINITY; j++) {
								SequenceItem tmp1 = best[j];
								best[j] = tmp;
								tmp = tmp1;
							}
							break;
						}
					}

				}
			}

			return best;
		}

		public SequencePrediction getBestSequences() {

			SequenceItem[] best = getBestSequencesAsItems();

			SequenceLabel[] d = new SequenceLabel[K];
			for (int k = 0; k < K; k++) {
				if (best[k].prob != Double.NEGATIVE_INFINITY) {
					d[k] = new SequenceLabel(getEntString(best[k]).split(" "),
							getFeatureVector(best[k]), best[k].prob);
				} else {
					d[k] = null;
				}
			}
			return new SequencePrediction(d);
		}

		public FeatureVector getFeatureVector(SequenceItem si) {
			if (si.prev == null)
				return si.fv;

			return FeatureVector.cat(getFeatureVector(si.prev), si.fv);
		}

		public String getEntString(SequenceItem si) {
			if (si.prev == null)
				return ((String) tagAlphabet.lookupObject(si.tagIdx)).trim();

			return (getEntString(si.prev) + " " + (String) tagAlphabet
					.lookupObject(si.tagIdx)).trim();
		}

	}

	public class SequenceItem {

		public int tokIdx, tagIdx;
		public double prob;
		public FeatureVector fv;
		public SequenceItem prev;

		// productions
		public SequenceItem(int tokIdx, int tagIdx, double prob,
				FeatureVector fv, SequenceItem prev) {

			this.tokIdx = tokIdx;
			this.tagIdx = tagIdx;
			this.prob = prob;
			this.fv = fv;
			this.prev = prev;
		}

		public SequenceItem() {
		}

	}

	public enum PredictCommitteeType {
		kBestViterbi, kDrawsByConst, kDrawsByStdDev
	};

	PredictCommitteeType committeeType;

	public static PredictCommitteeType resolveCommitteeType(
			String committeeTypeStr) {
		for (PredictCommitteeType type : PredictCommitteeType.values()) {
			if (type.toString().equalsIgnoreCase(committeeTypeStr)) {
				return type;
			}
		}
		throw new RuntimeException("Unkwon committee type: " + committeeTypeStr);
	}

	public void setCommitteeType(String committeeTypeStr) {
		PredictCommitteeType type = resolveCommitteeType(committeeTypeStr);
		setCommitteeType(type);
	}

	public void setCommitteeType(PredictCommitteeType committeeType) {
		if (this.committeeType != committeeType) {
			// Need to redraw the weight vectors...
			numOfGeneratedWeightVectors = 0;
			this.committeeType = committeeType;
		}
	}

	double drawsMultiplier = 1.0;

	public void setDrawsMultiplier(double multiplier) {
		if (drawsMultiplier != multiplier) {
			// Need to redraw the weight vectors...
			numOfGeneratedWeightVectors = 0;
			drawsMultiplier = multiplier;
		}
	}

	double[][] weightVectors;
	int numOfGeneratedWeightVectors;
	// Feature vector that hold the modified weights since last vectors draws
	FeatureVector changedWeightsVector;

	public void setModifiedWeightsVector(FeatureVector changedWeightsVector) {
		this.changedWeightsVector = changedWeightsVector;
	}

	// Draw K random weights vectors according to the mean and sigma values
	// The first weight vector is just the mean - the rest (K>1) contain changes
	// according to the sigma values, So when calculating the K best sequences
	// for K==1 this is done using the mean
	void drawKWeightsVectors(int K) {
		int i;

		if (K > numOfGeneratedWeightVectors) {
			weightVectors = new double[K][];
			// the very first weight vector is the mean
			weightVectors[0] = this.weights;
			numOfGeneratedWeightVectors = 1;

			// Use worker threads for drawing the weights vectors
			int numOfThreads = K - numOfGeneratedWeightVectors;
			DrawWeightThread[] drawers = new DrawWeightThread[numOfThreads];

			//
			// allocate the K output arrays + launch the threads
			//
			int trdIdx;
			for (i = numOfGeneratedWeightVectors, trdIdx = 0; i < K; i++, trdIdx++) {
				weightVectors[i] = new double[weights.length];
				drawers[trdIdx] = new DrawWeightThread(weightVectors[i]);
				drawers[trdIdx].start();
			}

			// Wait for the threads
			for (trdIdx = 0; trdIdx < drawers.length; trdIdx++) {
				try {
					drawers[trdIdx].join();
				} catch (InterruptedException e) {
				}
			}

			numOfGeneratedWeightVectors = i;
		}

		// If since last draw there was a model update (some weights were
		// updated) need to updated matching weights in all the drawn vectors
		if (changedWeightsVector != null) {
			FeatureVector curr;
			for (curr = changedWeightsVector; curr != null; curr = curr.next) {
				if (curr.index < 0)
					continue;
				int featureIndex = curr.index;
				for (int vectorIdx = 1; vectorIdx < numOfGeneratedWeightVectors; vectorIdx++) {
					weightVectors[vectorIdx][featureIndex] = drawWeight(random,
							featureIndex);
				}
			}
			// And set changed-weights vector to null
			changedWeightsVector = null;
		}
	}

	Random random = new Random();

	double drawWeight(Random rand, int idx) {
		double gauss = rand.nextGaussian();
		double stdDev = drawsMultiplier;
		if (committeeType == PredictCommitteeType.kDrawsByStdDev) {
			stdDev *= Math.sqrt(sigma[idx]);
		}
		double mean = weights[idx];
		double weight = mean + gauss * stdDev;
		return weight;
	}

	class DrawWeightThread extends Thread {
		double[] outputArr;
		Random rand;

		public DrawWeightThread(double[] outputArr) {
			this.outputArr = outputArr;
			this.rand = new Random();
		}

		public void run() {

			for (int i = 0; i < outputArr.length; i++) {
				// double gauss = rand.nextGaussian();
				// double stdDev = drawsMultiplier;
				// if (committeeType == PredictCommitteeType.kDrawsByStdDev) {
				// stdDev *= Math.sqrt(sigma[i]);
				// }
				// double mean = weights[i];
				// double weight = mean + gauss * stdDev;
				// outputArr[i] = weight;

				outputArr[i] = drawWeight(rand, i);
			}
		}
	}
}
