package struct.alg;

import struct.types.*;
import struct.sequence.*; 
import struct.solver.*; 

import java.io.*;

public class QPBatchUpdators implements BatchUpdator {

    // Parameters + Constructor
    static public enum UpdateType {MMMN_OPTIMAL};
    
    private UpdateType update_type;
    private double c;
    private Alphabet tagAlphabet;
    private int numLabels;
    
    private QPSolver qp_solver;

    public QPBatchUpdators(String params, Alphabet ta) {
	update_type = UpdateType.MMMN_OPTIMAL;
	c = Double.POSITIVE_INFINITY;
	QPParameters qp_params = new QPParameters();
	qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B00_FULL;
	
	tagAlphabet = ta;
	numLabels = tagAlphabet.size();
	
	// Decode Params
	System.out.println(params);
	String[] params_str = params.split(" ");
	
	System.out.println("Num of Params : " + params_str.length);
	
	for (int i=0; i<params_str.length; i++) {
	    params_str[i].trim();
	    String[] param = params_str[i].split(":");
	    
	    switch (param[0].charAt(0)) {
		
	    case 'u':
		if (param[1].compareTo("MMMN_OPTIMAL") == 0)
		    update_type = UpdateType.MMMN_OPTIMAL;
		else
		    throw new IllegalArgumentException("No such update type " + param[1]);
		break;
		
	    case 'C':
		c= Double.parseDouble(param[1].trim());
		break;
		
	    case 'n':
		qp_params.nu = Double.parseDouble(param[1].trim());
		break;


	    case 's':
		qp_params.eps = Double.parseDouble(param[1].trim());
		break;
		
	    default:
		System.out.println("Unkown option : " + params_str[i]);
	    } // switch
	} // for
	
	
	
	qp_solver = new QPSolver(qp_params);


	System.out.println(this.toString());
	System.out.println(qp_params.toString());
    }


    public String toString() {
	StringBuffer str = new StringBuffer();
	str.append("Update Type: " + update_type + " \n");
	str.append("C          : " + String.valueOf(c) + "\n");
	str.append("numLabels  : " + String.valueOf(numLabels) + "\n");
	return (str.toString());
    }


    public void set_c(double c1) {
	c=c1;
    }
        
    public UpdateType get_update_type() {
	return update_type;
    }

    public double get_c() {
	return c;
    }

    

    public void update(Instance[] inst, Predictor predictor, String featureFile) throws IOException {
	long start, end;

	start = System.currentTimeMillis();
	switch (update_type) {
	case MMMN_OPTIMAL:
	    mmmnOptimalUpdate(inst, predictor, featureFile);	    break;
	default:
	    System.out.println("No such update type " + update_type);
	}
	end = System.currentTimeMillis();
    }



    protected void mmmnOptimalUpdate(Instance[] instances, Predictor predictor, String featureFile) throws IOException {

	
	int n_constraints = 0;
	int n_messages = 0;
	Instance inst;
	Features feats;
	SequenceFeatures sequence_feats;
	SequenceInstance sequence_inst;
	SequenceLabel sequence_label;
	String[] actual_tags;
	int numTokens;

	
	
	// get total number of constraints
	for (int ind=0; ind<instances.length; ind++) {
	    inst = instances[ind];
	    
	    sequence_inst = (SequenceInstance) inst;
	    sequence_label = sequence_inst.getLabel();
	    actual_tags = sequence_label.tags;
	    numTokens = actual_tags.length; 

	    n_constraints += (numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1);
	    n_messages +=  2 * numLabels * numTokens;
	}
	System.out.println("total number of constraints = " + n_constraints);
	System.out.println("total number of messages    = " + n_messages);
	
	n_messages += instances.length; 
	
	// constraints structure
	QPConstraints constraints = new QPConstraints();
	constraints.a1 = new FeatureVector[n_constraints];
	constraints.a2 = new FeatureVector[n_constraints];
	constraints.b  = new double[n_constraints];
	
	constraints.dim1 = predictor.weights.length;
	constraints.dim2 = n_messages;
	constraints.C = c;

	// generate constraints;
	int alpha_base_index;// = 0;
	int beta_base_index;//  = alpha_base_index + numTokens * numLabels + 1;
	int global_base_index=0;

	ObjectInputStream in = new ObjectInputStream(new FileInputStream(featureFile));
	
	int i, k, k1, r;
	int act_ind, act_ind1;

	r = 0;
	for (int ind=0; ind<instances.length; ind++) {


	    inst  = instances[ind];
	    feats = instances[ind].getFeatures(in);
	    
	    sequence_inst = (SequenceInstance) inst;
	    sequence_label = sequence_inst.getLabel();
	    actual_tags = sequence_label.tags;
	    numTokens = actual_tags.length; // length of sequence
	    sequence_feats = (SequenceFeatures) feats;
	    
	    alpha_base_index = global_base_index;	    
	    beta_base_index  = alpha_base_index +  numTokens * numLabels + 1;

	    System.out.println(" ind = " + ind + 
			       " alpha_base_index " + alpha_base_index + 
			       " beta_base_index " +  beta_base_index + 
			       " global_base_index = " + global_base_index);

	
	    // first constraints
	    for (i=1; i<numTokens; i++) {
		act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
		for (k=0; k<numLabels-1; k++) {
		    
		    if (k == act_ind) continue;
		    
		    constraints.a1[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
								      sequence_feats.getFeatureVector(i, k) );
		    
		    constraints.b[r]  = 1;
		    
		    if (i > 0) { // forwoard only for i = 1 ... numTokens-1
			constraints.a2[r] = new FeatureVector ( i * numLabels + k + alpha_base_index, 1, null); // alpha_i[k<act_ind]
		    }
		    
		    if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
			
			if (constraints.a2 != null)
			    constraints.a2[r].add                 ( i * numLabels + k  + beta_base_index,  1); // -beta_i[k]
			else
			    constraints.a2[r] = new FeatureVector ( i * numLabels + k  + beta_base_index,  1, null); // -beta_i[k]
		    }
		    
		    r++;
		} // for k
	    } // for i
	    
	    // second constraints (going over all edges)
	    
	    for (i=1; i<numTokens-1; i++) {
		act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
		act_ind1 = tagAlphabet.lookupIndex(actual_tags[i+1]);
		
		for (k=0; k<numLabels-1; k++) { // k[i]
		    for (k1=0; k1<numLabels-1; k1++) { // k[i+1]
			if ( (k == act_ind) & (k1 == act_ind1) ) continue;
			
			constraints.a1[r] = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind, act_ind1) , 
									 sequence_feats.getFeatureVector(i, k, k1) );
			
			constraints.b[r] = 0;
			
			if (i > 0)
			    constraints.a2[r] = new FeatureVector ( (i+1)  * numLabels + k1  + alpha_base_index, -1, null); // +alpha_i[k]  
			if (i <numTokens-2)
			    if (constraints.a2[r] != null)
				constraints.a2[r].add                ( i  * numLabels + k  + beta_base_index, -1); // +beta_{i+1}[k1]
			    else
				constraints.a2[r] = new FeatureVector( i  * numLabels + k  + beta_base_index, -1, null); // +beta_{i+1}[k1]
			r++;
		    } // k1
		} //k
		
	    } // for i
	    
	    global_base_index += 2 *  numLabels * numTokens + 1;
	}
	System.out.println("r/n_cosntraints = " + r + "/" + n_constraints);
          
    

    }

}


