package struct.alg;

import struct.types.*;
import struct.sequence.*; 
import struct.solver.*; 

import java.util.Random;

public class QPUpdators implements OnlineUpdator {

    // Parameters + Constructor
    static public enum UpdateType {LLP_ALL, LLP_SAMPLE, LLP_L1, MMMN, MMMN_OPTIMAL};
    
    private UpdateType update_type;
    private double gamma;
    private double c;
    private double eta;
    private double rate;
    private Alphabet tagAlphabet;
    private int numLabels;
    
    private QPSolver qp_solver;

    public QPUpdators(String params, Alphabet ta) {
	update_type = UpdateType.LLP_ALL;
	c = Double.POSITIVE_INFINITY;
	gamma =1e0;
	eta = Double.POSITIVE_INFINITY;
	rate = 0.01;
	QPParameters qp_params = new QPParameters();
	
	tagAlphabet = ta;
	numLabels = tagAlphabet.size();
	
	// Decode Params
	System.out.println(params);
	String[] params_str = params.split(" ");
	
	System.out.println("Num of Params : " + params_str.length);
	
	for (int i=0; i<params_str.length; i++) {
	    params_str[i].trim();
	    String[] param = params_str[i].split(":");
	    
	    switch (param[0].charAt(0)) {
		
	    case 'u':
		if (param[1].compareTo("LLP_ALL") == 0)
		    update_type = UpdateType.LLP_ALL;
		else if (param[1].compareTo("LLP_L1") == 0)
		    update_type = UpdateType.LLP_L1;
		else if (param[1].compareTo("LLP_SAMPLE") == 0)
		    update_type = UpdateType.LLP_SAMPLE;
		else if (param[1].compareTo("MMMN") == 0)
		    update_type = UpdateType.MMMN;
		else if (param[1].compareTo("MMMN_OPTIMAL") == 0)
		    update_type = UpdateType.MMMN_OPTIMAL;
		else
		    throw new IllegalArgumentException("No such update type " + param[1]);
		break;
		
	    case 'C':
		c= Double.parseDouble(param[1].trim());
		break;
		
	    case 'g':
		gamma = Double.parseDouble(param[1].trim());
		break;
		
 	    case 'e':
		eta = Double.parseDouble(param[1].trim());
		break;

	    case 'r':
		rate = Double.parseDouble(param[1].trim());
		break;

	    case 'q':
		if (param[1].compareTo("HILDRETH") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH;
		else if (param[1].compareTo("HILDRETH_SEQ") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ;
		else if (param[1].compareTo("HILDRETH_SEQ_FAST") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_FAST;
		else if (param[1].compareTo("HILDRETH_SEQ_ALG_B0") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B0;
		else if (param[1].compareTo("HILDRETH_SEQ_ALG_B00") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B00;
		else if (param[1].compareTo("HILDRETH_SEQ_ALG_B00_BLAS") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B00_BLAS;
		else if (param[1].compareTo("HILDRETH_SEQ_ALG_B") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_B;
		else if (param[1].compareTo("HILDRETH_SEQ_ALG_A") == 0)
		    qp_params.qpsolver_type = QPParameters.QPSolverType.HILDRETH_SEQ_ALG_A;
		else
		    throw new IllegalArgumentException("No such qpsolver type " + param[1]);
		break;
		
	    case 'n':
		qp_params.nu = Double.parseDouble(param[1].trim());
		break;


	    case 's':
		qp_params.eps = Double.parseDouble(param[1].trim());
		break;
		
	    default:
		System.out.println("Unkown option : " + params_str[i]);
	    } // switch
	} // for
	
	
	
	qp_solver = new QPSolver(qp_params);


	System.out.println(this.toString());
	System.out.println(qp_params.toString());
    }


    public String toString() {
	StringBuffer str = new StringBuffer();
	str.append("Update Type: " + update_type + " \n");
	str.append("Gamma      : " + String.valueOf(gamma) + "\n");
	str.append("C          : " + String.valueOf(c) + "\n");
	str.append("Eta        : " + String.valueOf(eta) + "\n");
	str.append("numLabels  : " + String.valueOf(numLabels) + "\n");
	str.append("rate       : " + String.valueOf(rate) + "\n");
	return (str.toString());
    }


    public void set_gamma(double g) {
	gamma=g;
    }

    public void set_c(double c1) {
	c=c1;
    }


    public void set_eta(double eta1) {
	eta=eta1;
    }

        
    public UpdateType get_update_type() {
	return update_type;
    }

    public double get_gamma() {
	return gamma;
    }

    public double get_c() {
	return c;
    }


    public double get_eta() {
	return eta;
    }

    

    public void update(Instance inst, Features feats, Predictor predictor, double avg_upd) {

	// monitor
	Prediction pred = predictor.decode(inst,feats);
	Label currect_label = inst.getLabel();
	pred.getBestLabel();

	System.out.println("loss = " + inst.getLabel().loss(pred.getBestLabel()));
	System.out.println("curr = " + ((SequenceLabel)inst.getLabel()).toString());
	System.out.println("pred = " + ((SequenceLabel)pred.getBestLabel()).toString());


	QPConstraints constraints;

	long start, end;

	start = System.currentTimeMillis();
	switch (update_type) {
	case LLP_ALL:
	    constraints = llpUpdateAll(inst, feats, predictor);	    break;
	case LLP_SAMPLE:
	    constraints = llpUpdateSample(inst, feats, predictor);  break;
	case LLP_L1:
	    constraints = llpUpdateL1(inst, feats, predictor);	    break;
	case MMMN:
	    constraints = mmmnUpdate(inst, feats, predictor);	    break;
	case MMMN_OPTIMAL:
	    constraints = mmmnOptimalUpdate(inst, feats, predictor);	    break;
	default:
	    constraints = new QPConstraints();
	    System.out.println("No such update type " + update_type);
	}
	end = System.currentTimeMillis();
	System.out.println("Time to generate constraints: " + (end-start));


	
	double[] alpha = null;

	start = System.currentTimeMillis();
	alpha = qp_solver.solve(constraints);
	end = System.currentTimeMillis();
	System.out.println("Time run QP: " + (end-start));
			
	
	// update weights
	start = System.currentTimeMillis();
	updateWeights( constraints.a, alpha, avg_upd , predictor);
	end = System.currentTimeMillis();
	System.out.println("Time to update weights: " + (end-start));
	
    }

    protected QPConstraints llpUpdateAll(Instance inst, Features feats, Predictor predictor) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;

	// base of new features for slack
	int eta_base_index = predictor.weights.length + 2;

	// new feature
	double eta_slack_feature_value = Math.sqrt( numTokens / eta );
		
	// total number of constraints
	int n_constraints = (numTokens-1)*(numLabels-2)*(numLabels-1)*(numLabels-1);

	// print info
	System.out.println("Eta Base Index of Param " + eta_base_index);
	System.out.println("Number of Constraints = " + n_constraints);
	System.out.println("numTokens = " + numTokens);
	System.out.println("numLabels = " + numLabels);
	System.out.println("eta_slack_feature_value = " + eta_slack_feature_value);


	FeatureVector[] a = new FeatureVector[n_constraints];
	double[] b = new double[n_constraints];
	int i, k, kl,kr, r;
	int act_ind;
	FeatureVector el, er;
	
	r = 0;
	
	for (i=1; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels-1; k++) {
		
		if (k == act_ind) continue;
		
		for (kl=0; kl<numLabels-1; kl++) {
		    for (kr=0; kr<numLabels-1; kr++) {

			a[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
								    sequence_feats.getFeatureVector(i, k) );
			
			if (i > 0) { // also from left
			    //System.out.println("left:  i=" + i + " act_ind=" + act_ind + " k=" + k + " kl=" + kl + "  kr=" + kr);
			    el = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, kl, act_ind) , 
							      sequence_feats.getFeatureVector(i, kl, k) );
			    a[r] = FeatureVector.cat( a[r], el);
			}
			if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
			    //System.out.println("right: " + i + " " + act_ind + " " + k + " " + kl + "  " + kr);
			    er = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i+1, act_ind, kr) , 
							      sequence_feats.getFeatureVector(i+1, k, kr) );
			    a[r] = FeatureVector.cat( a[r], er);
			}
			b[r] = 1 - predictor.score(a[r]);
			a[r].add( eta_base_index + i , +eta_slack_feature_value ); 
			
			r++;
		    } // kr
		} // kl
	    } // for k
	} // for i
	System.out.println("r/n_constraints = " + r + "/" + n_constraints);

	
	QPConstraints constraints= new QPConstraints();
	constraints.a = a;
	constraints.b = b;

	return constraints;
    }




    protected QPConstraints llpUpdateSample(Instance inst, Features feats, Predictor predictor) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;

	// base of new features for slack
	int eta_base_index = predictor.weights.length + 2;

	// new feature
	double eta_slack_feature_value = Math.sqrt( numTokens / eta );
		
	// total number of constraints
	int n_constraints = numTokens*(numLabels-1)*numLabels*numLabels;

	// print info
	System.out.println("Eta Base Index of Param " + eta_base_index);
	System.out.println("Number of Constraints = " + n_constraints);
	System.out.println("numTokens = " + numTokens);
	System.out.println("numLabels = " + numLabels);
	System.out.println("eta_slack_feature_value = " + eta_slack_feature_value);


	Random rnd = new Random();
	boolean[] constraint_ind = new boolean[n_constraints];
	int actual_n_constraints=0;
	for (int r=0; r<n_constraints; r++)
	    if (rnd.nextDouble() > rate)
		constraint_ind[r] = false;
	    else {
		constraint_ind[r] = true;
		actual_n_constraints++;
	    }
	
	System.out.println("Actual Number of Constraints = " + actual_n_constraints);

	FeatureVector[] a = new FeatureVector[actual_n_constraints];
	double[] b = new double[actual_n_constraints];
	int i, k, kl,kr, r, r_actual;
	int act_ind;
	FeatureVector el, er;
	
	r = 0;
	r_actual = 0;
	
	for (i=0; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels; k++) {
		
		if (k == act_ind) continue;
		
		for (kl=0; kl<numLabels; kl++) {
		    for (kr=0; kr<numLabels; kr++) {

			if (constraint_ind[r] == true) {
			    a[r_actual] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
									sequence_feats.getFeatureVector(i, k) );
			    
			    if (i > 0) { // also from left
				//System.out.println("left:  i=" + i + " act_ind=" + act_ind + " k=" + k + " kl=" + kl + "  kr=" + kr);
				el = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, kl, act_ind) , 
								  sequence_feats.getFeatureVector(i, kl, k) );
				a[r_actual] = FeatureVector.cat( a[r_actual], el);
			    }
			    if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
				//System.out.println("right: " + i + " " + act_ind + " " + k + " " + kl + "  " + kr);
				er = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i+1, act_ind, kr) , 
								  sequence_feats.getFeatureVector(i+1, k, kr) );
				a[r_actual] = FeatureVector.cat( a[r_actual], er);
			    }
			    b[r_actual] = 1 - predictor.score(a[r_actual]);
			    a[r_actual].add( eta_base_index + i , +eta_slack_feature_value ); 
			    r_actual++;
			}
			r++;
			//System.out.println("r=" + r + " r_actual = " + r_actual);
		    } // kr
		} // kl
	    } // for k
	} // for i
	System.out.println("r/n_constraints = " + r + "/" + n_constraints);
	System.out.println("actual_r/actual_n_constraints = " + r_actual + "/" + actual_n_constraints);

	
	QPConstraints constraints = new QPConstraints();
	constraints.a = a;
	constraints.b = b;

	return constraints;
    }






    protected QPConstraints llpUpdateL1(Instance inst, Features feats, Predictor predictor) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;

	// base of new features for slack
	int eta_base_index = predictor.weights.length + 2;

	// new feature
	double eta_slack_feature_value = Math.sqrt( numTokens / eta );
		
	// total number of constraints
	int n_constraints = (numTokens-1)*(numLabels-1);

	// print info
	System.out.println("Eta Base Index of Param " + eta_base_index);
	System.out.println("Number of Constraints = " + n_constraints);
	System.out.println("numTokens = " + numTokens);
	System.out.println("numLabels = " + numLabels);
	System.out.println("eta_slack_feature_value = " + eta_slack_feature_value);

	FeatureVector[] a = new FeatureVector[n_constraints];
	double[] b = new double[n_constraints];
	int i, k, kl,kr, r;
	int act_ind;
	FeatureVector el, er;
	
	r = 0;
	
	for (i=1; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels; k++) {
		
		if (k == act_ind) continue;

		a[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
						     sequence_feats.getFeatureVector(i, k) );
		
		if (i > 0) { // also from left
		    kl = tagAlphabet.lookupIndex(actual_tags[i-1]);
		    el = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, kl, act_ind) , 
						      sequence_feats.getFeatureVector(i, kl, k) );
		    a[r] = FeatureVector.cat( a[r], el);
		}
		
		if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
		    kr = tagAlphabet.lookupIndex(actual_tags[i+1]);
		    er = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i+1, act_ind, kr) , 
						      sequence_feats.getFeatureVector(i+1, k, kr) );
		    a[r] = FeatureVector.cat( a[r], er);
		}
		b[r] = 1 - predictor.score(a[r]);
		a[r].add( eta_base_index + i , +eta_slack_feature_value ); 
		r++;
	    } // for k
	} // for i
	System.out.println("r/n_constraints = " + r + "/" + n_constraints);


	
	QPConstraints constraints = new QPConstraints();
	constraints.a = a;
	constraints.b = b;

	return constraints;
    }


    protected QPConstraints mmmnUpdate(Instance inst, Features feats, Predictor predictor) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;


	double gamma_slack_feature_value = Math.sqrt( 2*(numTokens-1)*numLabels) / gamma ;
	int alpha_base_index = predictor.weights.length + 2;
	int beta_base_index  = alpha_base_index + 2 * numTokens * numLabels + 10;
	System.out.println("Alpha Base Index of Param " + alpha_base_index);
	System.out.println("Beta  Base Index of Param " + beta_base_index);
	int n_cosntraints = (numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1);
	System.out.println("Number of Constraints (n_cosntraints) = " + n_cosntraints);
	System.out.println("gamma_slack_feature_value = " + gamma_slack_feature_value);
	
	FeatureVector[] a = new FeatureVector[n_cosntraints];
	double[] b = new double[n_cosntraints];
	int i, k, k1, r;
	int act_ind, act_ind1;

	
	r = 0;
	
	// first constraints
	for (i=1; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels-1; k++) {
		
		if (k == act_ind) continue;
		
		a[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
						     sequence_feats.getFeatureVector(i, k) );
		b[r] = 1 - predictor.score(a[r]);
		if (i > 0) { // forwoard only for i = 1 ... numTokens-1
		    a[r].add ( i * numLabels + k       + alpha_base_index, gamma_slack_feature_value); // alpha_i[k<act_ind]
		}
		if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
		    a[r].add ( i * numLabels + k       + beta_base_index,  gamma_slack_feature_value); // -beta_i[k]
		}
		r++;
	    } // for k
	} // for i
	
	// second constraints (going over all edges)
	FeatureVector temp1, temp2, temp3, temp4;
	for (i=1; i<numTokens-1; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    act_ind1 = tagAlphabet.lookupIndex(actual_tags[i+1]);
	    
	    for (k=0; k<numLabels-1; k++) { // k[i]
		for (k1=0; k1<numLabels-1; k1++) { // k[i+1]
		    if ( (k == act_ind) & (k1 == act_ind1) ) continue;
		    
		    temp1 = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind, act_ind1) , 
							 sequence_feats.getFeatureVector(i, k, k1) );
				
		    //temp2 = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind),		sequence_feats.getFeatureVector(i, k) );
		    //temp3 = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i+1, act_ind1) , sequence_feats.getFeatureVector(i+1, 1) );
		    //temp4 = FeatureVector.cat(temp1, temp2);
		    //a[r] =  FeatureVector.cat( temp3, temp4);
		    a[r] = temp1;
		    b[r] = - predictor.score(a[r]);
		    

		    // old code, buged !!
		    //if (i > 0)
		    //	a[r].add ( i  * numLabels + k  + alpha_base_index, gamma_slack_feature_value); // +alpha_i[k]
		    //if (i <numTokens-2)
		    //	a[r].add ( (i+1)  * numLabels + k1  + beta_base_index, gamma_slack_feature_value); // +beta_{i+1}[k1]
		    if (i > 0)
		    	a[r].add ( (i+1)  * numLabels + k1  + alpha_base_index, -gamma_slack_feature_value); // +alpha_i[k]
		    if (i <numTokens-2)
		    	a[r].add ( i  * numLabels + k  + beta_base_index, -gamma_slack_feature_value); // +beta_{i+1}[k1]
		    r++;
		} // k1
	    } //k
	    
	} // for i
	
	System.out.println("r/n_cosntraints = " + r + "/" + n_cosntraints);
        
    
    
	QPConstraints constraints = new QPConstraints();
	constraints.a = a;
	constraints.b = b;
	constraints.dim = beta_base_index + numLabels * numTokens;
	
	return constraints;
    }




    protected QPConstraints mmmnOptimalUpdate(Instance inst, Features feats, Predictor predictor) {

	// get needed information
	SequenceInstance sequence_inst = (SequenceInstance) inst;
	SequenceLabel sequence_label = sequence_inst.getLabel();
	String[] actual_tags = sequence_label.tags;
	int numTokens = actual_tags.length; // length of sequence
	SequenceFeatures sequence_feats = (SequenceFeatures) feats;

	int eta_base_index = predictor.weights.length + 2;
	double eta_slack_feature_value = Math.sqrt( 1 / eta );
		

	int alpha_base_index = 1;
	int beta_base_index  = alpha_base_index + numTokens * numLabels + 1;
	System.out.println("Alpha Base Index of Param " + alpha_base_index);
	System.out.println("Beta  Base Index of Param " + beta_base_index);
	int n_cosntraints = (numLabels-2)*(numTokens-1) + (numTokens-2)*((numLabels-1)*(numLabels-1)-1);
	System.out.println("Number of Constraints (n_cosntraints) = " + n_cosntraints);

	QPConstraints constraints = new QPConstraints();

	constraints.a1 = new FeatureVector[n_cosntraints];
	constraints.a2 = new FeatureVector[n_cosntraints];
	constraints.a  = constraints.a1;
	constraints.b  = new double[n_cosntraints];
	
	constraints.dim1 = eta_base_index + n_cosntraints;//predictor.weights.length ;
	constraints.dim2 = 2 * numLabels * numTokens + 2;

	constraints.C = c;
	
	int i, k, k1, r;
	int act_ind, act_ind1;

	
	r = 0;
	
	// first constraints
	for (i=1; i<numTokens; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    
	    for (k=0; k<numLabels-1; k++) {
		
		if (k == act_ind) continue;
		
		constraints.a1[r] =  FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind) , 
								  sequence_feats.getFeatureVector(i, k) );

		constraints.b[r]  = 1 - predictor.score(constraints.a1[r]);

		if (i > 0) { // forwoard only for i = 1 ... numTokens-1
		    constraints.a2[r] = new FeatureVector ( i * numLabels + k + alpha_base_index, 1, null); // alpha_i[k<act_ind]
		}

		if (i < numTokens - 1) { // backwards only for i = 0 ... numTokens-2
		
		    if (constraints.a2 != null)
			constraints.a2[r].add                 ( i * numLabels + k  + beta_base_index,  1); // -beta_i[k]
		    else
			constraints.a2[r] = new FeatureVector ( i * numLabels + k  + beta_base_index,  1, null); // -beta_i[k]
		}
		
		r++;
	    } // for k
	} // for i
	
	// second constraints (going over all edges)
	FeatureVector temp1, temp2, temp3, temp4;
	for (i=1; i<numTokens-1; i++) {
	    act_ind = tagAlphabet.lookupIndex(actual_tags[i]);
	    act_ind1 = tagAlphabet.lookupIndex(actual_tags[i+1]);
	    
	    for (k=0; k<numLabels-1; k++) { // k[i]
		for (k1=0; k1<numLabels-1; k1++) { // k[i+1]
		    if ( (k == act_ind) & (k1 == act_ind1) ) continue;
		    
		    constraints.a1[r] = FeatureVector.getDistVector( sequence_feats.getFeatureVector(i, act_ind, act_ind1) , 
								     sequence_feats.getFeatureVector(i, k, k1) );

		    constraints.b[r] = - predictor.score(constraints.a1[r]);
		    
		    if (i > 0)
			constraints.a2[r] = new FeatureVector ( (i+1)  * numLabels + k1  + alpha_base_index, -1, null); // +alpha_i[k]  
		    if (i <numTokens-2)
			if (constraints.a2[r] != null)
			    constraints.a2[r].add                ( i  * numLabels + k  + beta_base_index, -1); // +beta_{i+1}[k1]
			else
			    constraints.a2[r] = new FeatureVector( i  * numLabels + k  + beta_base_index, -1, null); // +beta_{i+1}[k1]
		    r++;
		} // k1
	    } //k
	    
	} // for i
	
	System.out.println("r/n_cosntraints = " + r + "/" + n_cosntraints);
        
    
    

	return constraints;
    }














    /////////////////////////////////////

    protected void updateWeights( FeatureVector[] constraints_fv,
				  double[] constraints_alpha,
				  double avg_upd,
				  Predictor predictor) {
	
	double[] weights = predictor.weights;
	double[] avg_weights = predictor.avg_weights;


	FeatureVector fv  = null;
	double alpha;
	
	if (constraints_fv.length != constraints_alpha.length) {
	    System.out.println("QPUpdators.updateWeights : dimension of constraints and dual variables do not match");
	    return;
	}
	
	
	
	for (int i=0; i < constraints_fv.length && constraints_fv[i] != null; i++) {
	    
	    fv     = constraints_fv[i];
	    alpha  = constraints_alpha[i];
	    
	    
	    if (fv.next == null)
		System.out.println("QPUpdators.updateWeights : empty FeatureVector");
	    
	    if (alpha != 0) {
		for (FeatureVector curr = fv; curr != null; curr = curr.next) {
		    
		    if (curr.index < 0) continue;
		    if (curr.index < weights.length) {
			weights[curr.index]     +=           alpha * curr.value;
			avg_weights[curr.index] += avg_upd * alpha * curr.value;
		    } 
		    else {
			System.out.println("updateWeights : index ( " + curr.index + " )  above top value " + weights.length) ;
		    }
		    

		}
	    } 
	} // for i
    } // updateWeights
}


