package struct.discontinuous;

import struct.alg.*;
import struct.types.*;

import java.io.*;

public class DiscontinuousPredictor extends Predictor {

    public double threshold;

    public DiscontinuousPredictor(int dimensions) {
	super(dimensions);
	threshold = 0.0;
    }
    public DiscontinuousPredictor(int dimensions, double threshold) {
	super(dimensions);
	this.threshold = threshold;
    }

    public void setThreshold(double d) {
	threshold = d;
    }
    
    public void saveModel(String file) throws Exception {
	System.out.print("Saving model ... ");
	ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file));
	out.writeObject(weights);
	out.writeObject(DiscontinuousInstance.getDataAlphabet());
	out.close();
	System.out.println("done.");
    }

    public void loadModel(String file) throws Exception {
	System.out.print("Loading model ... ");
	ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
	weights = (double[])in.readObject();
	DiscontinuousInstance.setDataAlphabet((Alphabet)in.readObject());
	in.close();
	System.out.println("done.");
    }

    public DiscontinuousPrediction decode(Instance inst, Features feats) {
	
	DiscontinuousInstance sinst = (DiscontinuousInstance)inst;

	String[] toks = sinst.getInput().sentence;

	return decode(inst,feats,toks.length);
    }

    public DiscontinuousPrediction decode(Instance inst, Features feats, int K) {
	return decode(inst, feats, K, false);
    }

    public DiscontinuousPrediction decode(Instance inst, Features feats, boolean onlyPos) {
	DiscontinuousInstance sinst = (DiscontinuousInstance)inst;

	String[] toks = sinst.getInput().sentence;
	//int K = onlyPos ? sinst.getLabel().fv.length : toks.length;
	int K = toks.length;

	return decode(inst,feats,K,onlyPos);
    }

    public DiscontinuousPrediction decode(Instance inst, Features feats, int K, boolean onlyPos) {

	DiscontinuousInstance sinst = (DiscontinuousInstance)inst;

	String[] toks = sinst.getInput().sentence;
	
	KBestDiscontinuous bs = new KBestDiscontinuous(sinst,K);
	bs.setOnlyPos(onlyPos);

	DiscontinuousFeatures sf = (DiscontinuousFeatures)feats;

	FeatureVector z = sf.getFeatureVector(0);
	bs.add(0,score(z),z,null,0);
	
	for(int n2 = 1; n2 < toks.length; n2++) {
	    for(int n1 = 0; n1 < n2; n1++) {
	    
		FeatureVector fv_ij = sf.getFeatureVector(n1,n2);
		double prob_ij = score(fv_ij);
		FeatureVector fv_i = sf.getFeatureVector(n2);
		double prob_i = score(fv_i);
		
		DiscontinuousItem[] items = bs.getItems(n1);
		    
		if(items == null)
		    continue;
		    
		int strt = 0;
		for(int i = 0; i < items.length; i++) {
		    if(items[i].prob == Double.NEGATIVE_INFINITY)
			continue;
		    strt = bs.add(n2,prob_ij+prob_i+items[i].prob,
				  FeatureVector.cat(fv_ij,fv_i),items[i],strt);
		    if(strt < 0)
			break;
		}
		
	    }
	}
	
	DiscontinuousPrediction ans = bs.getBestDiscontinuous();
	return ans;
    }

    public class KBestDiscontinuous {

	private DiscontinuousItem[][] chart;
	private int K;
	private int num_tags;
	private boolean onlyPos;
	public KBestDiscontinuous(DiscontinuousInstance inst, int K) {
	    this.K = K;
	    chart = new DiscontinuousItem[inst.getInput().sentence.length][K];
	    onlyPos = false;
	}

	public void setOnlyPos(boolean onlyPos) {
	    this.onlyPos = onlyPos;
	}

	public int add(int n, double prob, FeatureVector fv, DiscontinuousItem prev, int strt) {
		
	    if(chart[n][0] == null) {
		for(int i = 0; i < K; i++)
		    chart[n][i] = new DiscontinuousItem(n,Double.NEGATIVE_INFINITY,null,null);
	    }

	    if(chart[n][K-1].prob > prob)
		return -1;
			
	    for(int i = strt; i < K; i++) {
		if(chart[n][i].prob < prob) {
		    DiscontinuousItem tmp = chart[n][i];
		    chart[n][i] = new DiscontinuousItem(n,prob,fv,prev);
		    for(int j = i+1; j < K && tmp.prob != Double.NEGATIVE_INFINITY; j++) {
			DiscontinuousItem tmp1 = chart[n][j];
			chart[n][j] = tmp;
			tmp = tmp1;
		    }
		    return i+1 >= K ? -1 : i+1;
		}

	    }
	    return -1;
	}

	public double getProb(int n) {
	    return getProb(n,0);
	}

	public double getProb(int n, int i) {
	    if(chart[n][i] != null)
		return chart[n][i].prob;
	    return Double.NEGATIVE_INFINITY;
	}

	public double[] getProbs(int n) {
	    double[] result = new double[K];
	    for(int i = 0; i < K; i++)
		result[i] =
		    chart[n][i] != null ? chart[n][i].prob : Double.NEGATIVE_INFINITY;
	    return result;
	}

	public DiscontinuousItem getItem(int n) {
	    return getItem(n,0);
	}

	public DiscontinuousItem getItem(int n, int i) {
	    if(chart[n][i] != null)
		return chart[n][i];
	    return null;
	}

	public DiscontinuousItem[] getItems(int n) {
	    if(chart[n][0] != null)
		return chart[n];
	    else {
		System.out.println(n);
		System.exit(0);
	    }
	    return null;
	}

	public DiscontinuousPrediction getBestDiscontinuous() {
		
	    DiscontinuousItem[] best = new DiscontinuousItem[K];
	    for(int i = 0; i < K; i++) {
		best[i] = new DiscontinuousItem(-1,Double.NEGATIVE_INFINITY,null,null);
	    }
		
	    for(int k = 0; k < K; k++) {
		int n = chart.length-1;
		DiscontinuousItem cand = chart[n][k];
		if(onlyPos && cand.prob < threshold) continue;
		
		for(int i = 0; i < K; i++) {
		    if(best[i].prob < cand.prob) {
			DiscontinuousItem tmp = best[i];
			best[i] = cand;
			for(int j = i+1; j < K && tmp.prob != Double.NEGATIVE_INFINITY; j++) {
			    DiscontinuousItem tmp1 = best[j];
			    best[j] = tmp;
			    tmp = tmp1;
			}
			break;
		    }
		}
		
	    }
		
	    int num = 0;
	    for(int k = 0; k < K; k++) {
		if(best[k].prob != Double.NEGATIVE_INFINITY)
		    num++;
		else
		    break;
	    }

	    String[][] labels = new String[num][];
	    FeatureVector[] fvs = new FeatureVector[num];
	    for(int k = 0; k < num; k++) {
		labels[k] = getLabel(best[k],chart.length).split(" ");
		fvs[k] = getFeatureVector(best[k]);
	    }

	    return new DiscontinuousPrediction(new DiscontinuousLabel(labels,fvs));
	}

	public FeatureVector getFeatureVector(DiscontinuousItem si) {
	    if(si.prev == null)
		return si.fv;

	    return FeatureVector.cat(getFeatureVector(si.prev),si.fv);
	}

	public String getLabel(DiscontinuousItem si, int length) {
	    String[] indxs = getOnIndeces(si).split(" ");
	    String[] label = new String[length];
	    for(int i = 0; i < label.length; i++)
		label[i] = "O";
	    for(int i = 0; i < indxs.length; i++)
		label[Integer.parseInt(indxs[i])] = "I";
	    String res = "";
	    for(int i = 0; i < label.length; i++)
		res += label[i] + " ";
	    return res.trim();
	}
	
	private String getOnIndeces(DiscontinuousItem si) {
	    if(si.prev == null)
		return si.n+"";

		return (getOnIndeces(si.prev) + " " + si.n).trim();
	}
	
    }

    public class DiscontinuousItem {

	public int n;
	public double prob;
	public FeatureVector fv;
	public DiscontinuousItem prev;
	
	// productions
	public DiscontinuousItem(int n,
				 double prob, FeatureVector fv,
				 DiscontinuousItem prev) {
	    
	    this.n = n;
	    this.prob = prob;
	    this.fv = fv;
	    this.prev = prev;
	}
	
	public DiscontinuousItem() {}
	
    }
}
