package struct.sequence;

import struct.alg.*;
import struct.types.*;

import java.io.*;

public class SequencePredictor extends Predictor {

    public SequencePredictor(int dimensions) {
	super(dimensions);
    }

    public void saveModel(String file) throws Exception {
	System.out.print("Saving model ... ");
	ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file));
	out.writeObject(weights);
	out.writeObject(SequenceInstance.getDataAlphabet());
	out.writeObject(SequenceInstance.getTagAlphabet());
	out.close();
	System.out.println("done.");
    }

    public void loadModel(String file) throws Exception {
	System.out.print("Loading model ... ");
	ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
	weights = (double[])in.readObject();
	SequenceInstance.setDataAlphabet((Alphabet)in.readObject());
	SequenceInstance.setTagAlphabet((Alphabet)in.readObject());
	in.close();
	System.out.println("done.");
    }

    public SequencePrediction decode(Instance inst, Features feats) {
	return decode(inst,feats,1);
    }

    public SequencePrediction decode(Instance inst, Features feats, int K) {
	SequenceInstance sinst = (SequenceInstance)inst;

	Alphabet tagAlphabet = SequenceInstance.getTagAlphabet();

	String[] toks = sinst.getInput().sentence;
	
	KBestSequence bs = new KBestSequence(tagAlphabet,sinst,K);
	
	
	SequenceFeatures sf = (SequenceFeatures)feats;

	bs.add(0,tagAlphabet.lookupIndex("O"),0.0,new FeatureVector(-1,-1.0,null),null,0);
	for(int i = 0; i < tagAlphabet.size(); i++)
	    bs.add(0,i,Double.NEGATIVE_INFINITY,new FeatureVector(-1,-1.0,null),null,0);
	
	for(int n = 1; n < toks.length; n++) {
	    for(int e = 0; e < tagAlphabet.size(); e++) {
		
		for(int prev_e = 0; prev_e < tagAlphabet.size(); prev_e++) {
		    
		    FeatureVector fv_ij = sf.getFeatureVector(n,prev_e,e);
		    double prob_ij = score(fv_ij);
		    FeatureVector fv_i = sf.getFeatureVector(n,e);
		    double prob_i = score(fv_i);
		    
		    SequenceItem[] items = bs.getItems(n-1,prev_e);
		    
		    if(items == null)
			continue;
		    
		    int strt = 0;
		    for(int i = 0; i < items.length; i++) {
			if(items[i].prob == Double.NEGATIVE_INFINITY)
			    continue;
			strt = bs.add(n,e,prob_ij+prob_i+items[i].prob,
				      FeatureVector.cat(fv_ij,fv_i),items[i],strt);
			if(strt < 0)
			    break;
		    }
		    
		}
		
	    }
	}
	
	return bs.getBestSequences();
    }

    public class KBestSequence {

	private SequenceItem[][][] chart;
	private int K;
	private int num_tags;
	private Alphabet tagAlphabet;
	public KBestSequence(Alphabet tagAlphabet, SequenceInstance inst, int K) {
	    this.K = K;
	    chart = new SequenceItem[inst.getInput().sentence.length][tagAlphabet.size()][K];
	    num_tags = tagAlphabet.size();
	    this.tagAlphabet = tagAlphabet;
	}

	public int add(int n, int e, double prob, FeatureVector fv, SequenceItem prev, int strt) {
		
	    if(chart[n][e][0] == null) {
		for(int i = 0; i < K; i++)
		    chart[n][e][i] = new SequenceItem(n,e,Double.NEGATIVE_INFINITY,null,null);
	    }

	    if(chart[n][e][K-1].prob > prob)
		return -1;
			
	    for(int i = strt; i < K; i++) {
		if(chart[n][e][i].prob < prob) {
		    SequenceItem tmp = chart[n][e][i];
		    chart[n][e][i] = new SequenceItem(n,e,prob,fv,prev);
		    for(int j = i+1; j < K && tmp.prob != Double.NEGATIVE_INFINITY; j++) {
			SequenceItem tmp1 = chart[n][e][j];
			chart[n][e][j] = tmp;
			tmp = tmp1;
		    }
		    return i+1 >= K ? -1 : i+1;
		}

	    }
	    return -1;
	}

	public double getProb(int n, int e) {
	    return getProb(n,e,0);
	}

	public double getProb(int n, int e, int i) {
	    if(chart[n][e][i] != null)
		return chart[n][e][i].prob;
	    return Double.NEGATIVE_INFINITY;
	}

	public double[] getProbs(int n, int e) {
	    double[] result = new double[K];
	    for(int i = 0; i < K; i++)
		result[i] =
		    chart[n][e][i] != null ? chart[n][e][i].prob : Double.NEGATIVE_INFINITY;
	    return result;
	}

	public SequenceItem getItem(int n, int e) {
	    return getItem(n,e,0);
	}

	public SequenceItem getItem(int n, int e, int i) {
	    if(chart[n][e][i] != null)
		return chart[n][e][i];
	    return null;
	}

	public SequenceItem[] getItems(int n, int e) {
	    if(chart[n][e][0] != null)
		return chart[n][e];
	    else {
		System.out.println(n + " " + e);
		System.exit(0);
	    }
	    return null;
	}

	public SequencePrediction getBestSequences() {

	    int n = chart.length-1;
		
	    SequenceItem[] best = new SequenceItem[K];
	    for(int i = 0; i < K; i++) {
		best[i] = new SequenceItem(-1,-1,Double.NEGATIVE_INFINITY,null,null);
	    }
		
	    for(int e = 0; e < num_tags; e++) {
		for(int k = 0; k < K; k++) {
		    SequenceItem cand = chart[n][e][k];

		    for(int i = 0; i < K; i++) {
			if(best[i].prob < cand.prob) {
			    SequenceItem tmp = best[i];
			    best[i] = cand;
			    for(int j = i+1; j < K && tmp.prob != Double.NEGATIVE_INFINITY; j++) {
				SequenceItem tmp1 = best[j];
				best[j] = tmp;
				tmp = tmp1;
			    }
			    break;
			}
		    }
				
		}
	    }
		
				
	    SequenceLabel[] d = new SequenceLabel[K];
	    for(int k = 0; k < K; k++) {
		if(best[k].prob != Double.NEGATIVE_INFINITY) {
		    d[k] = new SequenceLabel(getEntString(best[k]).split(" "),getFeatureVector(best[k]));
		}
		else {
		    d[k] = null;
		}
	    }
	    return new SequencePrediction(d);
	}

	public FeatureVector getFeatureVector(SequenceItem si) {
	    if(si.prev == null)
		return si.fv;

	    return FeatureVector.cat(getFeatureVector(si.prev),si.fv);
	}

	public String getEntString(SequenceItem si) {
	    if(si.prev == null)
		return ((String)tagAlphabet.lookupObject(si.e)).trim();

	    return (getEntString(si.prev) + " " + (String)tagAlphabet.lookupObject(si.e)).trim();
	}
	
    }

    public class SequenceItem {

	public int n,e;
	public double prob;
	public FeatureVector fv;
	public SequenceItem prev;
	
	// productions
	public SequenceItem(int n, int e,
			    double prob, FeatureVector fv,
			    SequenceItem prev) {
	    
	    this.n = n;
	    this.e = e;
	    this.prob = prob;
	    this.fv = fv;
	    this.prev = prev;
	}
	
	public SequenceItem() {}
	
    }

}
